import equinox as eqx
from diffrax import CubicInterpolation, backward_hermite_coefficients
from jax import numpy as jnp
from jaxtyping import Array, Float


def check_make_safe_invariants(
    ts: Float[Array, " n"], ys: Float[Array, " n d"]
) -> tuple[Float[Array, " n"], Float[Array, " n d"]]:
    ts_nan = jnp.isnan(ts)
    ys_row_nan = jnp.all(jnp.isnan(ys), axis=-1)

    ts = eqx.error_if(ts, ts_nan != ys_row_nan, "NaN rows in ys must match NaNs in ts.")

    transitions = jnp.sum(ts_nan[:-1] != ts_nan[1:])
    edges_only = (transitions <= 1) | ((transitions == 2) & ts_nan[0] & ts_nan[-1])
    any_valid = jnp.any(~ts_nan)
    ts = eqx.error_if(
        ts,
        ~(edges_only & any_valid),
        "NaNs in ts may only appear at the edges; the non-NaN region must be a single contiguous block with at least one valid timestamp.",
    )

    increasing = jnp.where(ts_nan[1:] | ts_nan[:-1], True, ts[1:] > ts[:-1])
    ts = eqx.error_if(ts, ~increasing, "ts must be strictly increasing where not NaN.")
    return ts, ys


def make_monotonic_increasing(ts: Float[Array, " n"]) -> Float[Array, " n"]:
    """
    Replace edge NaNs while making ts strictly increasing.

    Assumes make_safe_invariants have been checked.
    """
    n = ts.shape[0]
    idxs = jnp.arange(n)
    nan_mask = jnp.isnan(ts)

    first_valid_idx = jnp.argmax(~nan_mask)
    first_val = ts[first_valid_idx]
    step_left = jnp.maximum(jnp.spacing(first_val), jnp.finfo(ts.dtype).eps)
    ts_filled = jnp.where(
        idxs < first_valid_idx,
        first_val - step_left * (first_valid_idx - idxs),
        ts,
    )

    last_valid_idx = n - 1 - jnp.argmax((~nan_mask)[::-1])
    last_val = ts[last_valid_idx]
    step_right = jnp.maximum(jnp.spacing(last_val), jnp.finfo(ts.dtype).eps)
    ts_filled = jnp.where(
        idxs > last_valid_idx,
        last_val + step_right * (idxs - last_valid_idx),
        ts_filled,
    )

    return ts_filled


def pad_ends(xs: Float[Array, " n d"]) -> Float[Array, " n d"]:
    """
    Replace NaN rows at the start and end of xs with the nearest valid row.

    Assumes make_safe_invariants have been checked.
    """
    n = xs.shape[0]
    idxs = jnp.arange(n)
    row_nan = jnp.all(jnp.isnan(xs), axis=-1)

    first_valid_idx = jnp.argmax(~row_nan)
    leading_mask = (idxs < first_valid_idx) & row_nan
    filled = jnp.where(
        leading_mask[:, None],
        jnp.broadcast_to(xs[first_valid_idx], xs.shape),
        xs,
    )

    last_valid_idx = n - 1 - jnp.argmax((~row_nan)[::-1])
    trailing_mask = (idxs > last_valid_idx) & row_nan
    filled = jnp.where(
        trailing_mask[:, None],
        jnp.broadcast_to(xs[last_valid_idx], xs.shape),
        filled,
    )
    return filled


def make_safe(
    ts: Float[Array, " n"], ys: Float[Array, " n d"]
) -> tuple[Float[Array, " n"], Float[Array, " n d"]]:
    ts, ys = check_make_safe_invariants(ts, ys)
    return make_monotonic_increasing(ts), pad_ends(ys)


def safe_control(
    ts: Float[Array, " n"], ys: Float[Array, " n d"]
) -> CubicInterpolation:
    ts, ys = make_safe(ts, ys)
    coeffs = backward_hermite_coefficients(ts, ys)
    return CubicInterpolation(ts, coeffs)



from __future__ import annotations

from typing import ClassVar

from jaxtyping import Array, Float, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.model import MLP
from lerax.space import AbstractSpace, Discrete

from .base_q import AbstractStatelessQPolicy


class MLPQPolicy[ObsType: Real[Array, "..."]](AbstractStatelessQPolicy[ObsType]):
    """
    Q-learning policy with an MLP Q-network.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        epsilon: The epsilon value for epsilon-greedy action selection.
        q_network: The MLP Q-network used for action value estimation.

    Args:
        env: The environment to create the policy for.
        epsilon: The epsilon value for epsilon-greedy action selection.
        width_size: The width of the hidden layers in the MLP.
        depth: The number of hidden layers in the MLP.
        key: JAX PRNG key for parameter initialization.

    Raises:
        ValueError: If the environment's action space is not Discrete.
    """

    name: ClassVar[str] = "MLPQPolicy"

    action_space: Discrete
    observation_space: AbstractSpace[ObsType]

    epsilon: float
    q_network: MLP

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, Integer[Array, ""], ObsType],
        *,
        epsilon: float = 0.1,
        width_size: int = 64,
        depth: int = 2,
        key: Key,
    ):
        if not isinstance(env.action_space, Discrete):
            raise ValueError(
                f"MLPQPolicy only supports Discrete action spaces, got {type(env.action_space)}"
            )

        self.action_space = env.action_space
        self.observation_space = env.observation_space

        self.epsilon = epsilon
        self.q_network = MLP(
            in_size=self.observation_space.flat_size,
            out_size=self.action_space.n,
            width_size=width_size,
            depth=depth,
            key=key,
        )

    def q_values(self, observation: ObsType) -> Float[Array, " actions"]:
        flat_obs = self.observation_space.flatten_sample(observation)
        return self.q_network(flat_obs)



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.render import (
    Abstract2DRenderer,
    AbstractRenderer,
    Color,
    PygameRenderer,
    Transform,
)
from lerax.space import Box, Discrete

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class AcrobotState(AbstractClassicControlEnvState):
    y: Float[Array, "4"]
    t: Float[Array, ""]


class Acrobot(
    AbstractClassicControlEnv[AcrobotState, Int[Array, ""], Float[Array, "4"]]
):
    """
    Acrobot environment matching the [Gymnasium Acrobot environment](https://gymnasium.farama.org/environments/classic_control/acrobot/).

    Note:
        To achieve identical dynamics to Gymnasium set `solver=diffrax.Euler()`.

    ## Action Space

    The action space is discrete with three actions:

    - 0: Apply -1 torque to the joint between the two links
    - 1: Apply 0 torque
    - 2: Apply +1 torque to the joint between the two links

    The action applies a fixed magnitude torque to the joint for the duration of the time step.

    ## Observation Space

    The observation space is a 6-dimensional continuous space representing the state of the two links:

    | Index | Observation               | Min Value | Max Value |
    |-------|---------------------------|-----------|-----------|
    | 0     | Cosine of Joint Angle 1   | -1.0      | 1.0       |
    | 1     | Sine of Joint Angle 1     | -1.0      | 1.0       |
    | 2     | Cosine of Joint Angle 2   | -1.0      | 1.0       |
    | 3     | Sine of Joint Angle 2     | -1.0      | 1.0       |
    | 4     | Joint Velocity 1          | -4π       | 4π        |
    | 5     | Joint Velocity 2          | -9π       | 9π        |

    ## Reward

    Non-terminal steps yield a reward of -1.0 and the terminal step yields a reward of 0.0.

    ## Termination

    The episode terminates when the tip of the second link reaches a height greater than 1.0 unit above the base.
    This corresponds to the condition: -cos(theta1) - cos(theta1 + theta2) > 1.0

    Args:
        gravity: Gravitational acceleration.
        link_length_1: Length of the first link.
        link_length_2: Length of the second link.
        link_mass_1: Mass of the first link.
        link_mass_2: Mass of the second link.
        link_com_pos_1: Center of mass position of the first link.
        link_com_pos_2: Center of mass position of the second link.
        link_moi: Moment of inertia of the links.
        max_vel_1: Maximum angular velocity for the first joint.
        max_vel_2: Maximum angular velocity for the second joint.
        torque_max_noise: Maximum noise to add to the applied torque.
        torques: Array of possible torques corresponding to each action.
        dt: Time step for integration.
        solver: Diffrax solver to use for integration.
        stepsize_controller: Step size controller for adaptive solvers.
    """

    name: ClassVar[str] = "Acrobot"

    action_space: Discrete
    observation_space: Box

    gravity: Float[Array, ""]
    link_length_1: Float[Array, ""]
    link_length_2: Float[Array, ""]
    link_mass_1: Float[Array, ""]
    link_mass_2: Float[Array, ""]
    link_com_pos_1: Float[Array, ""]
    link_com_pos_2: Float[Array, ""]
    link_moi: Float[Array, ""]
    max_vel_1: Float[Array, ""]
    max_vel_2: Float[Array, ""]
    torque_max_noise: Float[Array, ""]
    torques: Float[Array, "3"]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        gravity: Float[ArrayLike, ""] = 9.8,
        link_length_1: Float[ArrayLike, ""] = 1.0,
        link_length_2: Float[ArrayLike, ""] = 1.0,
        link_mass_1: Float[ArrayLike, ""] = 1.0,
        link_mass_2: Float[ArrayLike, ""] = 1.0,
        link_com_pos_1: Float[ArrayLike, ""] = 0.5,
        link_com_pos_2: Float[ArrayLike, ""] = 0.5,
        link_moi: Float[ArrayLike, ""] = 1.0,
        max_vel_1: Float[ArrayLike, ""] = 4 * jnp.pi,
        max_vel_2: Float[ArrayLike, ""] = 9 * jnp.pi,
        torque_max_noise: Float[ArrayLike, ""] = 0.0,
        torques: Float[ArrayLike, "3"] = jnp.array([-1.0, 0.0, 1.0]),
        dt: Float[ArrayLike, ""] = 0.2,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.gravity = jnp.array(gravity)
        self.link_length_1 = jnp.array(link_length_1)
        self.link_length_2 = jnp.array(link_length_2)
        self.link_mass_1 = jnp.array(link_mass_1)
        self.link_mass_2 = jnp.array(link_mass_2)
        self.link_com_pos_1 = jnp.array(link_com_pos_1)
        self.link_com_pos_2 = jnp.array(link_com_pos_2)
        self.link_moi = jnp.array(link_moi)
        self.max_vel_1 = jnp.array(max_vel_1)
        self.max_vel_2 = jnp.array(max_vel_2)
        self.torque_max_noise = jnp.array(torque_max_noise)
        self.torques = jnp.array(torques)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            if is_adaptive:
                self.stepsize_controller = diffrax.PIDController(rtol=1e-5, atol=1e-5)
            else:
                self.stepsize_controller = diffrax.ConstantStepSize()
        else:
            self.stepsize_controller = stepsize_controller

        self.action_space = Discrete(3)
        state_high = jnp.array([1.0, 1.0, 1.0, 1.0, self.max_vel_1, self.max_vel_2])
        low = -state_high
        self.observation_space = Box(low=low, high=state_high)

    def initial(self, *, key: Key) -> AcrobotState:
        return AcrobotState(
            y=jr.uniform(key, shape=(4,), minval=-0.1, maxval=0.1), t=jnp.array(0.0)
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "4"], action: Int[Array, ""]
    ) -> Float[Array, "4"]:
        theta1, theta2, theta1_d, theta2_d = y
        a = self.torques[action]

        d1 = (
            self.link_mass_1 * self.link_com_pos_1**2
            + self.link_mass_2
            * (
                self.link_length_1**2
                + self.link_com_pos_2**2
                + 2 * self.link_length_1 * self.link_com_pos_2 * jnp.cos(theta2)
            )
            + self.link_moi
            + self.link_moi
        )
        d2 = (
            self.link_mass_2
            * (
                self.link_com_pos_2**2
                + self.link_length_1 * self.link_com_pos_2 * jnp.cos(theta2)
            )
            + self.link_moi
        )

        phi2 = (
            self.link_mass_2
            * self.link_com_pos_2
            * self.gravity
            * jnp.cos(theta1 + theta2 - jnp.pi / 2)
        )
        phi1 = (
            -self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta2_d**2
            * jnp.sin(theta2)
            - 2
            * self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta1_d
            * theta2_d
            * jnp.sin(theta2)
            + (
                self.link_mass_1 * self.link_com_pos_1
                + self.link_mass_2 * self.link_length_1
            )
            * self.gravity
            * jnp.cos(theta1 - jnp.pi / 2)
            + phi2
        )

        theta2_dd = (
            a
            + d2 / d1 * phi1
            - self.link_mass_2
            * self.link_length_1
            * self.link_com_pos_2
            * theta1_d**2
            * jnp.sin(theta2)
            - phi2
        ) / (self.link_mass_2 * self.link_com_pos_2**2 + self.link_moi - d2**2 / d1)
        theta1_dd = -(d2 * theta2_dd + phi1) / d1
        return jnp.array([theta1_d, theta2_d, theta1_dd, theta2_dd])

    def clip(self, y: Float[Array, "4"]) -> Float[Array, "4"]:
        joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2 = y
        joint_angle_1 = (joint_angle_1 + jnp.pi) % (2 * jnp.pi) - jnp.pi
        joint_angle_2 = (joint_angle_2 + jnp.pi) % (2 * jnp.pi) - jnp.pi
        joint_vel_1 = jnp.clip(joint_vel_1, -self.max_vel_1, self.max_vel_1)
        joint_vel_2 = jnp.clip(joint_vel_2, -self.max_vel_2, self.max_vel_2)

        return jnp.array([joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2])

    def observation(self, state: AcrobotState, *, key: Key) -> Float[Array, "4"]:
        joint_angle_1, joint_angle_2, joint_vel_1, joint_vel_2 = state.y
        return jnp.array(
            [
                jnp.cos(joint_angle_1),
                jnp.sin(joint_angle_1),
                jnp.cos(joint_angle_2),
                jnp.sin(joint_angle_2),
                joint_vel_1,
                joint_vel_2,
            ]
        )

    def reward(
        self,
        state: AcrobotState,
        action: Int[Array, ""],
        next_state: AcrobotState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        joint_angle_1, joint_angle_2 = next_state.y[0], next_state.y[1]
        done_angle = (
            -jnp.cos(joint_angle_1) - jnp.cos(joint_angle_1 + joint_angle_2) > 1.0
        )
        return done_angle.astype(float) - 1.0

    def terminal(self, state: AcrobotState, *, key: Key) -> Bool[Array, ""]:
        joint_angle_1, joint_angle_2 = state.y[0], state.y[1]
        done_angle = (
            -jnp.cos(joint_angle_1) - jnp.cos(joint_angle_1 + joint_angle_2) > 1.0
        )
        return done_angle

    def render(self, state: AcrobotState, renderer: AbstractRenderer):
        if not isinstance(renderer, Abstract2DRenderer):
            raise TypeError("Acrobot environment requires an Abstract2DRenderer.")

        th1, th2 = state.y[0], state.y[1]

        base = jnp.array([0.0, 0.0])
        p1 = base + jnp.array(
            [self.link_length_1 * jnp.sin(th1), -self.link_length_1 * jnp.cos(th1)]
        )
        p2 = p1 + jnp.array(
            [
                self.link_length_2 * jnp.sin(th1 + th2),
                -self.link_length_2 * jnp.cos(th1 + th2),
            ]
        )

        link_w = 0.2
        joint_r = link_w / 2
        link_color = Color(0.0, 0.8, 0.8)
        joint_color = Color(0.8, 0.8, 0.0)
        goal_color = Color(0.0, 0.0, 0.0)

        renderer.clear()

        y_goal = jnp.array(1.0)
        renderer.draw_line(
            start=jnp.array([-3.0, y_goal]),
            end=jnp.array([3.0, y_goal]),
            color=goal_color,
            width=0.01,
        )

        renderer.draw_line(base, p1, color=link_color, width=link_w)
        renderer.draw_line(p1, p2, color=link_color, width=link_w)

        renderer.draw_circle(base, radius=joint_r, color=joint_color)
        renderer.draw_circle(p1, radius=joint_r, color=joint_color)

        renderer.draw()

    def default_renderer(self) -> Abstract2DRenderer:
        width, height = 800, 600
        transform = Transform(
            width=width,
            height=height,
            scale=140.0,
            offset=jnp.array([width / 2, height / 2]),
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Key

from .base_callback import (
    AbstractCallback,
    AbstractCallbackState,
    AbstractCallbackStepState,
    IterationContext,
    ResetContext,
    StepContext,
    TrainingContext,
)


class CallbackListStepState(AbstractCallbackStepState):
    """
    State for CallbackList callback step.

    Attributes:
        states: List of step states for each callback.
    """

    states: list[AbstractCallbackStepState]


class CallbackListState(AbstractCallbackState):
    """
    State for CallbackList callback.

    Attributes:
        states: List of states for each callback.
    """

    states: list[AbstractCallbackState]


class CallbackList(AbstractCallback[CallbackListState, CallbackListStepState]):
    """
    Callback that aggregates multiple callbacks and forwards calls to them.

    Attributes:
        callbacks: List of callbacks to aggregate.

    Args:
        callbacks: List of callbacks to aggregate.
    """

    callbacks: list[AbstractCallback]

    def __init__(self, callbacks: list[AbstractCallback]) -> None:
        self.callbacks = callbacks

    def reset(self, ctx: ResetContext, *, key: Key) -> CallbackListState:
        states = [
            callback.reset(ctx, key=key)
            for callback, key in zip(self.callbacks, jr.split(key, len(self.callbacks)))
        ]
        return CallbackListState(states=states)

    def step_reset(self, ctx: ResetContext, *, key: Key) -> CallbackListStepState:
        states = [
            callback.step_reset(ctx, key=key)
            for callback, key in zip(self.callbacks, jr.split(key, len(self.callbacks)))
        ]
        return CallbackListStepState(states=states)

    def on_step(self, ctx: StepContext, *, key: Key) -> CallbackListStepState:
        contexts = [
            eqx.tree_at(lambda c: c.state, ctx, state) for state in ctx.state.states
        ]
        new_states = [
            callback.on_step(ctx, key=key)
            for callback, ctx, key in zip(
                self.callbacks, contexts, jr.split(key, len(self.callbacks))
            )
        ]
        return CallbackListStepState(states=new_states)

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> CallbackListState:
        contexts = [
            eqx.tree_at(lambda c: (c.state, c.step_state), ctx, state)
            for state in zip(ctx.state.states, ctx.step_state.states)
        ]
        new_states = [
            callback.on_iteration(ctx, key=key)
            for callback, ctx, key in zip(
                self.callbacks, contexts, jr.split(key, len(self.callbacks))
            )
        ]
        return CallbackListState(states=new_states)

    def on_training_start(self, ctx: TrainingContext, *, key: Key) -> CallbackListState:
        contexts = [
            eqx.tree_at(lambda c: (c.state, c.step_state), ctx, state)
            for state in zip(ctx.state.states, ctx.step_state.states)
        ]
        new_states = [
            callback.on_training_start(ctx, key=key)
            for callback, ctx, key in zip(
                self.callbacks, contexts, jr.split(key, len(self.callbacks))
            )
        ]
        return CallbackListState(states=new_states)

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> CallbackListState:
        contexts = [
            eqx.tree_at(lambda c: (c.state, c.step_state), ctx, state)
            for state in zip(ctx.state.states, ctx.step_state.states)
        ]
        new_states = [
            callback.on_training_end(ctx, key=key)
            for callback, ctx, key in zip(
                self.callbacks, contexts, jr.split(key, len(self.callbacks))
            )
        ]
        return CallbackListState(states=new_states)

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        contexts = [
            eqx.tree_at(lambda c: (c.state, c.step_state), ctx, state)
            for state in zip(ctx.state.states, ctx.step_state.states)
        ]
        continue_flags = [
            callback.continue_training(ctx, key=key)
            for callback, ctx, key in zip(
                self.callbacks, contexts, jr.split(key, len(self.callbacks))
            )
        ]
        return jnp.all(jnp.array(continue_flags))



from __future__ import annotations

from distreqx import distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float

from .base_distribution import AbstractDistreqxWrapper, AbstractMaskableDistribution


class Bernoulli(
    AbstractMaskableDistribution[Bool[Array, " dims"]],
    AbstractDistreqxWrapper[Bool[Array, " dims"]],
):
    """
    Bernoulli distribution.

    Attributes:
        distribution: The underlying distreqx Bernoulli distribution.

    Args:
        logits: The log-odds of the distribution.
        probs: The probabilities of the distribution.
    """

    distribution: distributions.Bernoulli

    def __init__(
        self,
        logits: Float[ArrayLike, " dims"] | None = None,
        probs: Float[ArrayLike, " dims"] | None = None,
    ):
        logits = jnp.asarray(logits) if logits is not None else None
        probs = jnp.asarray(probs) if probs is not None else None
        self.distribution = distributions.Bernoulli(logits=logits, probs=probs)

    @property
    def logits(self) -> Float[Array, " dims"]:
        return self.distribution.logits

    @property
    def probs(self) -> Float[Array, " dims"]:
        return self.distribution.probs

    def mask(self, mask: Bool[Array, " dims"]) -> Bernoulli:
        masked_logits = jnp.where(mask, self.logits, -jnp.inf)
        return Bernoulli(logits=masked_logits)



from __future__ import annotations

from typing import Callable, NamedTuple

from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float

from lerax.space import Box


class RescaleResult(NamedTuple):
    box: Box
    forward: Callable[[Float[ArrayLike, " ..."]], Float[Array, " ..."]]
    backward: Callable[[Float[ArrayLike, " ..."]], Float[Array, " ..."]]


def rescale_box(
    box: Box, min: Float[ArrayLike, " ..."], max: Float[ArrayLike, " ..."]
) -> RescaleResult:
    """
    Return functions to rescale one box space to another.

    Logic stolen from Gymnasium library.

    Returns:
        A rescaled box, a function to map from the original to the rescaled box, and a
        function to map from the rescaled box to the original.
    """

    min = jnp.broadcast_to(min, box.shape)
    max = jnp.broadcast_to(max, box.shape)

    assert jnp.all(min <= max)
    assert jnp.all(box.low <= box.high)

    min_finite = jnp.isfinite(min)
    max_finite = jnp.isfinite(max)
    both_finite = min_finite & max_finite

    gradient = jnp.ones_like(min)
    gradient = gradient.at[both_finite].set(
        (max[both_finite] - min[both_finite])
        / (box.high[both_finite] - box.low[both_finite])
    )

    intercept = jnp.zeros_like(min)
    intercept = intercept.at[max_finite].set(max[max_finite] - box.high[max_finite])
    intercept = intercept.at[min_finite].set(
        min[min_finite] - box.low[min_finite] * gradient[min_finite]
    )

    new_box = Box(low=min, high=max, shape=box.shape)

    def forward(sample: Float[ArrayLike, " ..."]) -> Float[Array, " ..."]:
        sample = jnp.asarray(sample)
        return gradient * sample + intercept

    def backward(sample: Float[ArrayLike, " ..."]) -> Float[Array, " ..."]:
        sample = jnp.asarray(sample)
        return (sample - intercept) / gradient

    return RescaleResult(new_box, forward, backward)



from __future__ import annotations

import operator
from functools import reduce
from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from .base_space import AbstractSpace
from .utils import try_cast


class MultiBinary(AbstractSpace[Bool[Array, " n"]]):
    """
    A space of binary values.

    All dimensions are independent and can take values 0 or 1.

    Attributes:
        n: The shape of the multi-binary space.

    Args:
        n: The shape of the multi-binary space. If an integer is provided,
            then the shape will be (n,).
    """

    n: tuple[int, ...]

    def __init__(self, n: int | tuple[int, ...]):
        if isinstance(n, int):
            assert n > 0, "n must be positive"
            self.n = (n,)
        else:
            assert all(
                isinstance(dim, int) and dim > 0 for dim in n
            ), "all dimensions in n must be positive integers"
            self.n = n

    @property
    def shape(self) -> tuple[int, ...]:
        return self.n

    def canonical(self) -> Bool[Array, " n"]:
        return jnp.zeros(self.shape, dtype=bool)

    def sample(self, key: Key) -> Bool[Array, " n"]:
        return jr.bernoulli(key, shape=self.shape)

    def contains(self, x: Any) -> Bool[Array, ""]:
        x = try_cast(x)
        if x is None:
            return jnp.array(False)

        if x.shape != self.shape:
            return jnp.array(False)

        return jnp.all((x == 0) | (x == 1), axis=0)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiBinary):
            return False

        return self.n == other.n

    def __repr__(self) -> str:
        return f"MultiBinary({self.n})"

    def __hash__(self) -> int:
        return hash(self.n)

    def flatten_sample(self, sample: Bool[ArrayLike, " n"]) -> Float[Array, " n"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> int:
        return reduce(operator.mul, self.n)



from __future__ import annotations

import equinox as eqx
import jax
import optax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Key, Scalar

from lerax.buffer import RolloutBuffer
from lerax.policy import AbstractStatefulActorCriticPolicy
from lerax.utils import filter_scan

from .on_policy import AbstractOnPolicyAlgorithm, OnPolicyState, OnPolicyStepState


class PPOStats(eqx.Module):
    """
    PPO training statistics.

    Attributes:
        approx_kl: Approximate KL divergence between old and new policy.
        total_loss: Total loss.
        policy_loss: Policy loss.
        value_loss: Value function loss.
        entropy_loss: Entropy loss.
    """

    approx_kl: Float[Array, ""]
    total_loss: Float[Array, ""]
    policy_loss: Float[Array, ""]
    value_loss: Float[Array, ""]
    entropy_loss: Float[Array, ""]


class PPO[PolicyType: AbstractStatefulActorCriticPolicy](
    AbstractOnPolicyAlgorithm[PolicyType]
):
    """
    Proximal Policy Optimization (PPO) algorithm.

    Attributes:
        optimizer: The optimizer used for training.
        gae_lambda: Lambda parameter for Generalized Advantage Estimation (GAE).
        gamma: Discount factor.
        num_envs: Number of parallel environments.
        num_steps: Number of steps to run for each environment per update.
        batch_size: Size of each training batch.
        num_epochs: Number of epochs to train the policy per update.
        normalize_advantages: Whether to normalize advantages.
        clip_coefficient: Clipping coefficient for policy and value function updates.
        clip_value_loss: Whether to clip the value function loss.
        entropy_loss_coefficient: Coefficient for the entropy loss term.
        value_loss_coefficient: Coefficient for the value function loss term.
        max_grad_norm: Maximum gradient norm for gradient clipping.

    Args:
        num_envs: Number of parallel environments.
        num_steps: Number of steps to run for each environment per update.
        num_epochs: Number of epochs to train the policy per update.
        num_batches: Number of batches to split the rollout buffer into for training.
        gae_lambda: Lambda parameter for Generalized Advantage Estimation (GAE).
        gamma: Discount factor.
        clip_coefficient: Clipping coefficient for policy and value function updates.
        clip_value_loss: Whether to clip the value function loss.
        entropy_loss_coefficient: Coefficient for the entropy loss term.
        value_loss_coefficient: Coefficient for the value function loss term.
        max_grad_norm: Maximum gradient norm for gradient clipping.
        normalize_advantages: Whether to normalize advantages.
        learning_rate: Learning rate for the optimizer.
    """

    optimizer: optax.GradientTransformation

    gae_lambda: float
    gamma: float

    num_envs: int
    num_steps: int
    batch_size: int
    num_epochs: int

    normalize_advantages: bool
    clip_coefficient: float
    clip_value_loss: bool
    entropy_loss_coefficient: float
    value_loss_coefficient: float
    max_grad_norm: float

    def __init__(
        self,
        *,
        num_envs: int = 4,
        num_steps: int = 512,
        num_epochs: int = 16,
        num_batches: int = 32,
        gae_lambda: float = 0.95,
        gamma: float = 0.99,
        clip_coefficient: float = 0.2,
        clip_value_loss: bool = False,
        entropy_loss_coefficient: float = 0.0,
        value_loss_coefficient: float = 0.5,
        max_grad_norm: float = 0.5,
        normalize_advantages: bool = True,
        learning_rate: optax.ScalarOrSchedule = 3e-4,
    ):
        self.gae_lambda = gae_lambda
        self.gamma = gamma

        self.num_envs = num_envs
        self.num_steps = num_steps
        self.num_epochs = num_epochs
        self.batch_size = (self.num_steps * self.num_envs) // num_batches

        self.clip_coefficient = clip_coefficient
        self.clip_value_loss = clip_value_loss
        self.entropy_loss_coefficient = entropy_loss_coefficient
        self.value_loss_coefficient = value_loss_coefficient
        self.max_grad_norm = max_grad_norm
        self.normalize_advantages = normalize_advantages

        adam = optax.inject_hyperparams(optax.adam)(learning_rate)
        clip = optax.clip_by_global_norm(self.max_grad_norm)
        self.optimizer = optax.chain(clip, adam)

    def per_step(
        self, step_state: OnPolicyStepState[PolicyType]
    ) -> OnPolicyStepState[PolicyType]:
        return step_state

    def per_iteration(
        self, state: OnPolicyState[PolicyType]
    ) -> OnPolicyState[PolicyType]:
        return state

    # Needs to be static so the first argument can be a policy
    # eqx.filter_value_and_grad doesn't support argnums
    @staticmethod
    def ppo_loss(
        policy: PolicyType,
        rollout_buffer: RolloutBuffer,
        normalize_advantages: bool,
        clip_coefficient: float,
        clip_value_loss: bool,
        value_loss_coefficient: float,
        entropy_loss_coefficient: float,
    ) -> tuple[Float[Array, ""], PPOStats]:
        _, values, log_probs, entropy = jax.vmap(policy.evaluate_action)(
            rollout_buffer.states, rollout_buffer.observations, rollout_buffer.actions
        )

        values = eqx.error_if(values, ~jnp.isfinite(values), "Non-finite values.")
        log_probs = eqx.error_if(
            log_probs, ~jnp.isfinite(log_probs), "Non-finite log_probs."
        )
        entropy = eqx.error_if(entropy, ~jnp.isfinite(entropy), "Non-finite entropy.")

        log_ratios = log_probs - rollout_buffer.log_probs
        ratios = jnp.exp(log_ratios)
        approx_kl = jnp.mean(ratios - log_ratios) - 1

        advantages = rollout_buffer.advantages
        if normalize_advantages:
            advantages = (advantages - jnp.mean(advantages)) / (
                jnp.std(advantages) + jnp.finfo(advantages.dtype).eps
            )

        policy_loss = -jnp.mean(
            jnp.minimum(
                advantages * ratios,
                advantages
                * jnp.clip(ratios, 1 - clip_coefficient, 1 + clip_coefficient),
            )
        )

        if clip_value_loss:
            clipped_values = rollout_buffer.values + jnp.clip(
                values - rollout_buffer.values, -clip_coefficient, clip_coefficient
            )
            value_loss = (
                jnp.mean(
                    jnp.minimum(
                        jnp.square(values - rollout_buffer.returns),
                        jnp.square(clipped_values - rollout_buffer.returns),
                    )
                )
                / 2
            )
        else:
            value_loss = jnp.mean(jnp.square(values - rollout_buffer.returns)) / 2

        entropy_loss = -jnp.mean(entropy)

        loss = (
            policy_loss
            + value_loss * value_loss_coefficient
            + entropy_loss * entropy_loss_coefficient
        )

        return loss, PPOStats(
            approx_kl,
            loss,
            policy_loss,
            value_loss,
            entropy_loss,
        )

    ppo_loss_grad = staticmethod(eqx.filter_value_and_grad(ppo_loss, has_aux=True))

    def train_batch(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        rollout_buffer: RolloutBuffer,
    ) -> tuple[PolicyType, optax.OptState, PPOStats]:
        (_, stats), grads = self.ppo_loss_grad(
            policy,
            rollout_buffer,
            self.normalize_advantages,
            self.clip_coefficient,
            self.clip_value_loss,
            self.value_loss_coefficient,
            self.entropy_loss_coefficient,
        )

        updates, new_opt_state = self.optimizer.update(
            grads, opt_state, eqx.filter(policy, eqx.is_inexact_array)
        )
        policy = eqx.apply_updates(policy, updates)

        return policy, new_opt_state, stats

    def train_epoch(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        rollout_buffer: RolloutBuffer,
        *,
        key: Key,
    ) -> tuple[PolicyType, optax.OptState, PPOStats]:
        def batch_scan(
            carry: tuple[
                PolicyType,
                optax.OptState,
            ],
            buffer: RolloutBuffer,
        ):
            policy, opt_state = carry
            policy, opt_state, stats = self.train_batch(policy, opt_state, buffer)
            return (policy, opt_state), stats

        (policy, opt_state), stats = filter_scan(
            batch_scan,
            (policy, opt_state),
            rollout_buffer.batches(self.batch_size, key=key),
        )
        stats = jax.tree.map(jnp.mean, stats)
        return policy, opt_state, stats

    @staticmethod
    def explained_variance(
        returns: Float[Array, ""], values: Float[Array, ""]
    ) -> Float[Array, ""]:
        variance = jnp.var(returns)
        return 1 - jnp.var(returns - values) / (variance + jnp.finfo(returns.dtype).eps)

    def train(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        buffer: RolloutBuffer,
        *,
        key: Key,
    ) -> tuple[PolicyType, optax.OptState, dict[str, Scalar]]:
        def epoch_scan(
            carry: tuple[PolicyType, optax.OptState], key: Key
        ) -> tuple[tuple[PolicyType, optax.OptState], PPOStats]:
            policy, opt_state = carry
            policy, opt_state, stats = self.train_epoch(
                policy, opt_state, buffer, key=key
            )
            return (policy, opt_state), stats

        (policy, opt_state), stats = filter_scan(
            epoch_scan, (policy, opt_state), jr.split(key, self.num_epochs)
        )

        stats = jax.tree.map(jnp.mean, stats)
        explained_variance = self.explained_variance(buffer.returns, buffer.values)
        log = {
            "approx_kl": stats.approx_kl,
            "loss": stats.total_loss,
            "policy_loss": stats.policy_loss,
            "value_loss": stats.value_loss,
            "entropy_loss": stats.entropy_loss,
            "explained_variance": explained_variance,
        }
        return policy, opt_state, log



from .base_actor_critic import (
    AbstractActorCriticPolicy,
    AbstractStatefulActorCriticPolicy,
    AbstractStatelessActorCriticPolicy,
    ActorCriticStatefulWrapper,
)
from .mlp import MLPActorCriticPolicy
from .ncde import NCDEActorCriticPolicy

__all__ = [
    "AbstractActorCriticPolicy",
    "AbstractStatelessActorCriticPolicy",
    "AbstractStatefulActorCriticPolicy",
    "ActorCriticStatefulWrapper",
    "MLPActorCriticPolicy",
    "NCDEActorCriticPolicy",
]



from __future__ import annotations

from typing import ClassVar

from jax import random as jr
from jaxtyping import Array, Float, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.model import MLP, ActionLayer
from lerax.space import AbstractSpace

from .base_actor_critic import AbstractStatelessActorCriticPolicy


class MLPActorCriticPolicy[
    ActType: (Float[Array, " dims"], Integer[Array, ""]),
    ObsType: Real[Array, "..."],
](AbstractStatelessActorCriticPolicy[ActType, ObsType]):
    """
    Actor–critic policy with MLP components.

    Uses an MLP to encode observations into features, then separate MLPs to
    produce action distributions and value estimates from those features.

    Action distributions are produced by mapping action head outputs to the
    parameters of the appropriate distribution for the action space.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        encoder: MLP to encode observations into features.
        value_head: MLP to produce value estimates from features.
        action_head: MLP to produce action distributions from features.

    Args:
        env: The environment to create the policy for.
        feature_size: Size of the feature representation.
        feature_width: Width of the hidden layers in the feature encoder.
        feature_depth: Depth of the hidden layers in the feature encoder.
        value_width: Width of the hidden layers in the value head.
        value_depth: Depth of the hidden layers in the value head.
        action_width: Width of the hidden layers in the action head.
        action_depth: Depth of the hidden layers in the action head.
        log_std_init: Initial log standard deviation for continuous action spaces.
        key: JAX PRNG key for parameter initialization.
    """

    name: ClassVar[str] = "MLPActorCriticPolicy"

    action_space: AbstractSpace[ActType]
    observation_space: AbstractSpace[ObsType]

    encoder: MLP
    value_head: MLP
    action_head: ActionLayer

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        *,
        feature_size: int = 16,
        feature_width: int = 64,
        feature_depth: int = 2,
        value_width: int = 64,
        value_depth: int = 2,
        action_width: int = 64,
        action_depth: int = 2,
        log_std_init: float = 0.0,
        key: Key,
    ):
        self.action_space = env.action_space
        self.observation_space = env.observation_space

        feat_key, val_key, act_key = jr.split(key, 3)

        self.encoder = MLP(
            in_size=self.observation_space.flat_size,
            out_size=feature_size,
            width_size=feature_width,
            depth=feature_depth,
            key=feat_key,
        )

        self.value_head = MLP(
            in_size=feature_size,
            out_size="scalar",
            width_size=value_width,
            depth=value_depth,
            key=val_key,
        )

        self.action_head = ActionLayer(
            self.action_space,
            feature_size,
            action_width,
            action_depth,
            key=act_key,
            log_std_init=log_std_init,
        )

    def __call__(self, observation: ObsType, *, key: Key | None = None) -> ActType:
        features = self.encoder(self.observation_space.flatten_sample(observation))
        action_dist = self.action_head(features)

        if key is None:
            action = action_dist.mode()
        else:
            action = action_dist.sample(key)

        return action

    def action_and_value(
        self, observation: ObsType, *, key: Key
    ) -> tuple[ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """
        features = self.encoder(self.observation_space.flatten_sample(observation))
        value = self.value_head(features)

        action_dist = self.action_head(features)
        action, log_prob = action_dist.sample_and_log_prob(key)

        return action, value, log_prob.sum().squeeze()

    def value(self, observation: ObsType) -> Float[Array, ""]:
        """Get the value of an observation."""
        features = self.encoder(self.observation_space.flatten_sample(observation))
        return self.value_head(features)

    def evaluate_action(
        self, observation: ObsType, action: ActType
    ) -> tuple[Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """Evaluate an action given an observation."""
        features = self.encoder(self.observation_space.flatten_sample(observation))
        action_dist = self.action_head(features)
        value = self.value_head(features)
        log_prob = action_dist.log_prob(action)

        try:
            entropy = action_dist.entropy().squeeze()
        except NotImplementedError:
            entropy = -log_prob.mean().squeeze()  # Fallback to negative log prob mean

        return value, log_prob.sum().squeeze(), entropy



from .base_dqn import (
    AbstractDQNPolicy,
    AbstractStatefulDQNPolicy,
    AbstractStatelessDQNPolicy,
    DQNStatefulWrapper,
)
from .mlp import MLPDQNPolicy

__all__ = [
    "AbstractDQNPolicy",
    "AbstractStatelessDQNPolicy",
    "AbstractStatefulDQNPolicy",
    "DQNStatefulWrapper",
    "MLPDQNPolicy",
]



from __future__ import annotations

from typing import ClassVar, Literal

try:
    import gymnasium as gym
except ModuleNotFoundError as e:
    raise ImportError(
        "Gymnasium compatibility requires optional dependencies. "
        "Install with `pip install lerax[compatibility]` or install "
        "`gymnasium` manually."
    ) from e
import jax
import numpy as np
from jax import numpy as jnp
from jax import random as jr
from jax.debug import callback as debug_callback
from jax.experimental import io_callback
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnv, AbstractEnvState
from lerax.render import AbstractRenderer
from lerax.space import (
    AbstractSpace,
    Box,
    Dict,
    Discrete,
    MultiBinary,
    MultiDiscrete,
    Tuple,
)


def gym_space_to_lerax_space(space: gym.Space) -> AbstractSpace:
    """
    Returns a Lerax space corresponding to the given Gymnasium space.

    Args:
        space: Gymnasium space to convert.

    Returns:
        The corresponding Lerax space.
    """
    if isinstance(space, gym.spaces.Discrete):
        if not space.start == 0:
            raise NotImplementedError(
                "Gym Discrete space with non-zero start are not supported"
            )
        return Discrete(n=int(space.n))
    elif isinstance(space, gym.spaces.Box):
        return Box(low=space.low, high=space.high, shape=space.shape)
    elif isinstance(space, gym.spaces.Dict):
        return Dict({k: gym_space_to_lerax_space(s) for k, s in space.spaces.items()})
    elif isinstance(space, gym.spaces.Tuple):
        return Tuple(tuple(gym_space_to_lerax_space(s) for s in space.spaces))
    elif isinstance(space, gym.spaces.MultiBinary):
        return MultiBinary(n=space.n)
    elif isinstance(space, gym.spaces.MultiDiscrete):
        return MultiDiscrete(ns=tuple(int(n) for n in space.nvec))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def lerax_to_gym_space(space: AbstractSpace) -> gym.Space:
    """
    Returns a Gymnasium space corresponding to the given Lerax space.

    Args:
        space: Lerax space to convert.

    Returns:
        The corresponding Gymnasium space.
    """

    if isinstance(space, Discrete):
        return gym.spaces.Discrete(int(space.n))
    elif isinstance(space, Box):
        return gym.spaces.Box(
            low=np.asarray(space.low),
            high=np.asarray(space.high),
        )
    elif isinstance(space, Dict):
        return gym.spaces.Dict(
            {k: lerax_to_gym_space(s) for k, s in space.spaces.items()}
        )
    elif isinstance(space, Tuple):
        return gym.spaces.Tuple(tuple(lerax_to_gym_space(s) for s in space.spaces))
    elif isinstance(space, MultiBinary):
        return gym.spaces.MultiBinary(
            n=int(space.n[0]) if len(space.n) == 1 else space.n
        )
    elif isinstance(space, MultiDiscrete):
        return gym.spaces.MultiDiscrete(nvec=list(space.ns))
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def jax_to_numpy(x):
    if isinstance(x, jnp.ndarray):
        return np.asarray(x)
    return x


def to_numpy_tree(x):
    return jax.tree.map(jax_to_numpy, x)


class GymEnvState(AbstractEnvState):
    observation: Array
    reward: Float[Array, ""]
    terminal: Bool[Array, ""]
    truncated: Bool[Array, ""]


class GymToLeraxEnv(AbstractEnv[GymEnvState, Array, Array]):
    """
    Wrapper of a Gymnasium environment to make it compatible with Lerax.

    Note:
        Uses jax's `io_callback` to wrap the env's `reset` and `step` functions.
        In general, this will be slower than a native JAX environment and prevents
        vmapped rollout. Also removes the info dict returned by Gymnasium envs since
        the shape cannot be known ahead of time. Even more so than normal it is
        important to only call methods in order since the state objects do not
        contain all necessary information.

    Args:
        env: Gymnasium environment to wrap.

    Attributes:
        name: Name of the environment.
        action_space: Action space of the environment.
        observation_space: Observation space of the environment.
        env: The original Gymnasium environment.
    """

    name: ClassVar[str] = "GymnasiumEnv"

    action_space: AbstractSpace
    observation_space: AbstractSpace

    env: gym.Env

    def __init__(self, env: gym.Env):
        self.env = env
        self.action_space = gym_space_to_lerax_space(env.action_space)
        self.observation_space = gym_space_to_lerax_space(env.observation_space)

    def initial(self, *args, key: Key, **kwargs) -> GymEnvState:
        """
        Forwards to the Gymnasium reset.

        Note:
            A seed is generated if none is provided to increase reproducibility.

        Args:
            *args: Positional arguments to pass to `env.reset`.
            key: JAX PRNG key, used to generate a seed if none is provided.
            **kwargs: Keyword arguments to pass to `env.reset`. If "seed" is
                provided here, it overrides the key-generated seed.

        Returns:
            The initial environment state.
        """
        if "seed" in kwargs:
            kwargs = dict(kwargs)
            seed_value = kwargs.pop("seed")
            seed = jnp.asarray(seed_value, dtype=int)
        else:
            seed = jr.randint(key, (), 0, jnp.iinfo(jnp.int32).max)

        def reset_callback(seed_arr):
            seed_int = int(seed_arr)
            obs, _ = self.env.reset(*args, seed=seed_int, **kwargs)
            return jnp.asarray(obs)

        observation = io_callback(
            reset_callback,
            self.observation_space.canonical(),
            seed,
            ordered=True,
        )

        return GymEnvState(
            observation=observation,
            reward=jnp.array(0.0, dtype=float),
            terminal=jnp.array(False, dtype=bool),
            truncated=jnp.array(False, dtype=bool),
        )

    def transition(self, state: GymEnvState, action: Array, *, key: Key) -> GymEnvState:
        """
        Forwards to the Gymnasium step.

        In practice, this just calls the env's step function via io_callback.
        This means that the state is ignored and order of operations is important.

        Args:
            state: Current environment state.
            action: Action to take.
            key: Unused.

        Returns:
            Next environment state.
        """

        def step_callback(action_arr):
            observation, reward, terminated, truncated, _ = self.env.step(
                np.asarray(action_arr)
            )
            return (
                jnp.asarray(observation),
                jnp.asarray(reward, dtype=float),
                jnp.asarray(terminated, dtype=bool),
                jnp.asarray(truncated, dtype=bool),
            )

        observation, reward, terminated, truncated = io_callback(
            step_callback,
            (
                self.observation_space.canonical(),
                jnp.array(0.0, dtype=float),
                jnp.array(False, dtype=bool),
                jnp.array(False, dtype=bool),
            ),
            action,
            ordered=True,
        )

        return GymEnvState(
            observation=observation,
            reward=reward,
            terminal=terminated,
            truncated=truncated,
        )

    def observation(self, state: GymEnvState, *, key: Key) -> Array:
        """
        Forwards to the Gymnasium observation.

        Args:
            state: Current environment state.

        Returns:
            Observation corresponding to the environment state.
        """
        return state.observation

    def reward(
        self, state: GymEnvState, action: Array, next_state: GymEnvState, *, key: Key
    ) -> Float[Array, ""]:
        """
        Forwards to the Gymnasium reward.

        In practice, this just reads the reward from the next_state.

        Args:
            state: Current environment state.
            action: Action taken.
            next_state: Next environment state.

        Returns:
            Reward obtained from the transition.
        """
        return next_state.reward

    def terminal(self, state: GymEnvState, *, key: Key) -> Bool[Array, ""]:
        """
        Forwards to the Gymnasium terminated flag.

        Args:
            state: Current environment state.

        Returns:
            Boolean indicating whether the state is terminal.
        """
        return state.terminal

    def truncate(self, state: GymEnvState) -> Bool[Array, ""]:
        """
        Forwards to the Gymnasium truncated flag.

        Args:
            state: Current environment state.

        Returns:
            Boolean indicating whether the state is truncated.
        """
        return state.truncated

    def state_info(self, state: GymEnvState) -> dict:
        """
        Empty info dict to ensure stable shapes for JIT compilation.

        Args:
            state: Current environment state.

        Returns:
            Empty info dict.
        """
        return {}

    def transition_info(
        self, state: GymEnvState, action: Array, next_state: GymEnvState
    ) -> dict:
        """
        Empty info dict to ensure stable shapes for JIT compilation.

        Args:
            state: Current environment state.
            action: Action taken.
            next_state: Next environment state.

        Returns:
            Empty info dict.
        """
        return {}

    def render(self, state: GymEnvState, renderer: AbstractRenderer):
        """
        Not supported for Gymnasium environments.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError("Rendering not implemented for GymToLeraxEnv")

    def default_renderer(self) -> AbstractRenderer:
        """
        Not supported for Gymnasium environments.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError("Default renderer not implemented for GymToLeraxEnv.")

    def close(self):
        debug_callback(self.env.close, ordered=True)


class LeraxToGymEnv[StateType: AbstractEnvState](gym.Env):
    """
    Wrapper of an Lerax environment to make it compatible with Gymnasium.

    Executes the Lerax env directly (Python side). Keeps an internal eqx state and PRNG.

    Attributes:
        metadata: Metadata for the Gym environment.
        action_space: Action space of the environment.
        observation_space: Observation space of the environment.
        render_mode: Render mode for the environment.
        env: The Lerax environment to wrap.
        state: Current state of the Lerax environment.
        key: PRNG key for the environment.

    Args:
        env: Lerax environment to wrap.
        render_mode: Render mode for the environment.
    """

    metadata: dict = {"render_modes": ["human"]}

    action_space: gym.Space
    observation_space: gym.Space

    render_mode: str | None = None

    env: AbstractEnv[StateType, Array, Array]
    state: StateType
    key: Key

    def __init__(
        self,
        env: AbstractEnv[StateType, Array, Array],
        render_mode: Literal["human"] | None = None,
    ):
        self.key = jr.key(0)

        self.env = env

        self.action_space = lerax_to_gym_space(env.action_space)
        self.observation_space = lerax_to_gym_space(env.observation_space)

        self.render_mode = render_mode
        # TODO: Actually handle rendering

    def reset(self, *, seed: int | None = None, options: dict | None = None):
        if seed is not None:
            self.key = jr.key(int(seed))

        self.key, reset_key = jr.split(self.key)
        self.state, obs, info = self.env.reset(key=reset_key)
        return jax_to_numpy(obs), to_numpy_tree(info)

    def step(self, action):
        self.key, step_key = jr.split(self.key)
        self.state, obs, rew, term, trunc, info = self.env.step(
            self.state, jnp.asarray(action), key=step_key
        )

        return (
            jax_to_numpy(obs),
            float(jnp.asarray(rew)),
            bool(jnp.asarray(term)),
            bool(jnp.asarray(trunc)),
            to_numpy_tree(info),
        )

    def render(self):
        """
        Not supported yet.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError("Rendering not implemented for LeraxToGymEnv")

    def close(self):
        """
        Placeholder close method.

        Does nothing but completes the Gymnasium Env interface.
        """



from __future__ import annotations

from functools import partial
from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.space import AbstractSpace, Box

from .base_wrapper import AbstractWrapper
from .utils import rescale_box


class AbstractPureObservationWrapper[
    WrapperObsType, StateType: AbstractEnvLikeState, ActType, ObsType
](AbstractWrapper[StateType, ActType, WrapperObsType, StateType, ActType, ObsType]):
    """
    Apply a pure function to every observation that leaves the environment.
    """

    env: eqx.AbstractVar[AbstractEnvLike[StateType, ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[ObsType], WrapperObsType]]
    observation_space: eqx.AbstractVar[AbstractSpace[WrapperObsType]]

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.env.action_space

    def initial(self, *, key: Key) -> StateType:
        return self.env.initial(key=key)

    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        return self.env.transition(state, action, key=key)

    def observation(self, state: StateType, *, key: Key) -> WrapperObsType:
        return self.func(self.env.observation(state, key=key))

    def reward(
        self, state: StateType, action: ActType, next_state: StateType, *, key: Key
    ) -> Float[Array, ""]:
        return self.env.reward(state, action, next_state, key=key)

    def terminal(self, state: StateType, *, key: Key) -> Bool[Array, ""]:
        return self.env.terminal(state, key=key)

    def truncate(self, state: StateType) -> Bool[Array, ""]:
        return self.env.truncate(state)

    def state_info(self, state: StateType) -> dict:
        return self.env.state_info(state)

    def transition_info(
        self, state: StateType, action: ActType, next_state: StateType
    ) -> dict:
        return self.env.transition_info(state, action, next_state)


class TransformObservation[
    WrapperObsType, StateType: AbstractEnvLikeState, ActType, ObsType
](AbstractPureObservationWrapper[WrapperObsType, StateType, ActType, ObsType]):
    """
    Apply an arbitrary function to every observation that leaves the environment.

    Attributes:
        env: The environment to wrap.
        observation_space: The observation space of the wrapper.

    Args:
        env: The environment to wrap.
        observation_space: The observation space of the wrapper.
    """

    env: AbstractEnvLike[StateType, ActType, ObsType]
    func: Callable[[ObsType], WrapperObsType]
    observation_space: AbstractSpace[WrapperObsType]

    def __init__(
        self,
        env: AbstractEnvLike[StateType, ActType, ObsType],
        func: Callable[[ObsType], WrapperObsType],
        observation_space: AbstractSpace[WrapperObsType],
    ):
        self.env = env
        self.func = func
        self.observation_space = observation_space


class ClipObservation[StateType: AbstractEnvLikeState](
    AbstractPureObservationWrapper[
        Float[Array, " ..."], StateType, Float[Array, " ..."], Float[Array, " ..."]
    ],
):
    """
    Clips every observation to the environment's observation space.

    Note:
        Only works with `Box` observation spaces.

    Attributes:
        env: The environment to wrap.
        observation_space: The observation space of the wrapper.

    Args:
        env: The environment to wrap.

    Raises:
        ValueError: If the environment's observation space is not a `Box`.
    """

    env: AbstractEnvLike
    func: Callable
    observation_space: Box

    def __init__(self, env: AbstractEnvLike):
        if not isinstance(env.observation_space, Box):
            raise ValueError(
                "ClipObservation only supports `Box` observation spaces"
                f" not {type(env.observation_space)}"
            )

        self.env = env
        self.func = partial(
            jnp.clip,
            min=env.observation_space.low,
            max=env.observation_space.high,
        )
        self.observation_space = env.observation_space


class RescaleObservation[StateType: AbstractEnvLikeState](
    AbstractPureObservationWrapper[
        Float[Array, " ..."], StateType, Float[Array, " ..."], Float[Array, " ..."]
    ],
):
    """
    Affine rescaling of the box observation space to a specified range.

    Attributes:
        env: The environment to wrap.
        observation_space: The observation space of the wrapper.

    Args:
        env: The environment to wrap.
        min: The minimum value of the rescaled observation space.
        max: The maximum value of the rescaled observation space.

    Raises:
        ValueError: If the environment's observation space is not a `Box`.
    """

    env: AbstractEnvLike
    func: Callable
    observation_space: Box

    def __init__(
        self,
        env: AbstractEnvLike,
        min: Float[Array, " ..."] = jnp.array(-1.0),
        max: Float[Array, " ..."] = jnp.array(1.0),
    ):
        if not isinstance(env.observation_space, Box):
            raise ValueError(
                "RescaleObservation only supports `Box` observation spaces"
                f" not {type(env.action_space)}"
            )

        new_box, forward, _ = rescale_box(env.observation_space, min, max)

        self.env = env
        self.func = forward
        self.observation_space = new_box


class FlattenObservation[StateType: AbstractEnvLikeState, ObsType](
    AbstractPureObservationWrapper[
        Float[Array, " flat"], StateType, Float[Array, " ..."], ObsType
    ]
):
    """
    Flatten the observation space into a 1-D array.

    Attributes:
        env: The environment to wrap.
        observation_space: The observation space of the wrapper.

    Args:
        env: The environment to wrap.
    """

    env: AbstractEnvLike
    func: Callable
    observation_space: Box

    def __init__(self, env: AbstractEnvLike):
        self.env = env
        self.func = self.env.observation_space.flatten_sample
        self.observation_space = Box(
            -jnp.inf,
            jnp.inf,
            shape=(int(jnp.asarray(self.env.observation_space.flat_size)),),
        )



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import optax
from jax import numpy as jnp
from jaxtyping import Array, Bool, Float, Int, Key

from lerax.env import AbstractEnvLike
from lerax.policy import AbstractPolicy


class AbstractCallbackStepState(eqx.Module):
    """
    Base class for callback states that are vectorized across environment
    steps.
    """


class AbstractCallbackState(eqx.Module):
    """Base class for callback states."""


class ResetContext(eqx.Module):
    """Context passed to the reset method of callbacks."""

    locals: dict


class StepContext[StepStateType: AbstractCallbackStepState](eqx.Module):
    """
    Values passed to step-related callback methods.

    Attributes:
        state: The current callback step state.
        env: The environment being interacted with.
        policy: The policy being used to interact with the environment.
        done: Boolean indicating if the episode has terminated or truncated.
        reward: Reward received from the environment at the current step.
        locals: A dictionary for storing additional information.
    """

    state: StepStateType
    env: AbstractEnvLike
    policy: AbstractPolicy
    done: Bool[Array, ""]
    reward: Float[Array, ""]
    locals: dict


class IterationContext[
    StateType: AbstractCallbackState, StepStateType: AbstractCallbackStepState
](eqx.Module):
    """
    Values passed to iteration-related callback methods.

    Attributes:
        state: The current callback state.
        step_state: The current callback step state.
        env: The environment being interacted with.
        policy: The policy being used to interact with the environment.
        iteration_count: The current training iteration count.
        opt_state: The current optimizer state.
        training_log: A dictionary containing training metrics.
        locals: A dictionary for storing additional information.
    """

    state: StateType
    step_state: StepStateType
    env: AbstractEnvLike
    policy: AbstractPolicy
    iteration_count: Int[Array, ""]
    opt_state: optax.OptState
    training_log: dict[str, Array]
    locals: dict


class TrainingContext[
    StateType: AbstractCallbackState, StepStateType: AbstractCallbackStepState
](eqx.Module):
    """
    Values passed to training-related callback methods.

    Attributes:
        state: The current callback state.
        step_state: The current callback step state.
        env: The environment being interacted with.
        policy: The policy being used to interact with the environment.
        total_timesteps: Total number of timesteps for training.
        iteration_count: The current training iteration count.
        locals: A dictionary for storing additional information.
    """

    state: StateType
    step_state: StepStateType
    env: AbstractEnvLike
    policy: AbstractPolicy
    total_timesteps: int
    iteration_count: Int[Array, ""]
    opt_state: optax.OptState
    locals: dict


class AbstractCallback[
    StateType: AbstractCallbackState, StepStateType: AbstractCallbackStepState
](eqx.Module):
    """
    Base class for RL algorithm callbacks.

    Note:
        All concrete methods should work under JIT compilation.
    """

    @abstractmethod
    def reset(self, ctx: ResetContext, *, key: Key) -> StateType:
        """Initialize the callback state."""

    @abstractmethod
    def step_reset(self, ctx: ResetContext, *, key: Key) -> StepStateType:
        """Reset the callback state for vectorized steps."""

    @abstractmethod
    def on_step(self, ctx: StepContext, *, key: Key) -> StepStateType:
        """Called at the end of each environment step."""

    @abstractmethod
    def on_iteration(self, ctx: IterationContext, *, key: Key) -> StateType:
        """Called at the end of each training iteration."""

    @abstractmethod
    def on_training_start(self, ctx: TrainingContext, *, key: Key) -> StateType:
        """Called at the start of training."""

    @abstractmethod
    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> StateType:
        """Called at the end of training."""

    @abstractmethod
    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        """Called at the end of each iteration to determine whether to continue."""


class EmptyCallbackStepState(AbstractCallbackStepState):
    """Empty step state for stateless callbacks."""


class EmptyCallbackState(AbstractCallbackState):
    """Empty state for stateless callbacks."""


class AbstractStatelessCallback(
    AbstractCallback[EmptyCallbackState, EmptyCallbackStepState]
):
    """Callback that does not maintain any state."""

    def reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackState:
        return EmptyCallbackState()

    def step_reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackStepState:
        return EmptyCallbackStepState()


class AbstractStepCallback[StepStateType: AbstractCallbackStepState](
    AbstractCallback[EmptyCallbackState, StepStateType]
):
    """Callback that only implements step-related methods."""

    def reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackState:
        return EmptyCallbackState()

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> EmptyCallbackState:
        return ctx.state

    def on_training_start(
        self, ctx: TrainingContext, *, key: Key
    ) -> EmptyCallbackState:
        return ctx.state

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> EmptyCallbackState:
        return ctx.state

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(True)


class AbstractIterationCallback[StateType: AbstractCallbackState](
    AbstractCallback[StateType, EmptyCallbackStepState]
):
    """Callback that only implements iteration-related methods."""

    def step_reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackStepState:
        return EmptyCallbackStepState()

    def on_step(self, ctx: StepContext, *, key: Key) -> EmptyCallbackStepState:
        return ctx.state

    def on_training_start(self, ctx: TrainingContext, *, key: Key) -> StateType:
        return ctx.state

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> StateType:
        return ctx.state

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(True)


class AbstractTrainingCallback[StateType: AbstractCallbackState](
    AbstractCallback[StateType, EmptyCallbackStepState]
):
    """Callback that only implements training-related methods."""

    def step_reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackStepState:
        return EmptyCallbackStepState()

    def on_step(self, ctx: StepContext, *, key: Key) -> EmptyCallbackStepState:
        return ctx.state

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> StateType:
        return ctx.state

    def continue_training(self, ctx: IterationContext, *, key: Key) -> Bool[Array, ""]:
        return jnp.array(True)


class AbstractContinueTrainingCallback[StateType: AbstractCallbackState](
    AbstractCallback[StateType, EmptyCallbackStepState]
):
    """Callback that only implements continue training method."""

    def step_reset(self, ctx: ResetContext, *, key: Key) -> EmptyCallbackStepState:
        return EmptyCallbackStepState()

    def on_step(self, ctx: StepContext, *, key: Key) -> EmptyCallbackStepState:
        return ctx.state

    def on_iteration(self, ctx: IterationContext, *, key: Key) -> StateType:
        return ctx.state

    def on_training_start(self, ctx: TrainingContext, *, key: Key) -> StateType:
        return ctx.state

    def on_training_end(self, ctx: TrainingContext, *, key: Key) -> StateType:
        return ctx.state



from __future__ import annotations

import contextlib
import warnings
from typing import Any

from jax import numpy as jnp
from jaxtyping import ArrayLike, Float

from .base_renderer import WHITE, Abstract2DRenderer, Color, Transform

pygame: Any
gfxdraw: Any


def _load_pygame() -> tuple[Any, Any]:
    """
    Import pygame and gfxdraw lazily.

    Raises a clear ImportError if the optional dependency is missing.
    """
    global pygame, gfxdraw

    if pygame is not None and gfxdraw is not None:
        return pygame, gfxdraw

    try:
        with contextlib.redirect_stdout(None):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=Warning)
                import pygame as _pygame
                from pygame import gfxdraw as _gfxdraw
    except ImportError as exc:
        raise ImportError(
            "PygameRenderer requires the optional rendering dependencies. "
            "Install them with: pip install lerax[render]"
        ) from exc

    pygame = _pygame
    gfxdraw = _gfxdraw
    return pygame, gfxdraw


class PygameRenderer(Abstract2DRenderer):
    """
    PyGame renderer implementation.

    Attributes:
        transform: Transform from world coordinates to screen pixels.
        width: The width of the rendering window in pixels.
        height: The height of the rendering window in pixels.
        screen: The PyGame surface representing the rendering window.
        background_color: The background color of the rendering window.

    Args:
        width: The width of the rendering window in pixels.
        height: The height of the rendering window in pixels.
        background_color: The background color of the rendering window.
        transform: Transform from world coordinates to screen pixels.
    """

    transform: Transform

    width: int
    height: int

    screen: Any
    background_color: Color

    def __init__(
        self,
        width: int,
        height: int,
        background_color: Color = WHITE,
        transform: Transform | None = None,
    ) -> None:
        pg, _ = _load_pygame()

        self.width = width
        self.height = height
        self.background_color = background_color

        if transform is None:
            self.transform = Transform(
                width=width,
                height=height,
                scale=jnp.array(1.0),
                offset=jnp.array([0.0, 0.0]),
            )
        else:
            self.transform = transform

        pg.init()
        self.screen = pg.display.set_mode((self.width, self.height))

    def open(self):
        pass

    def close(self):
        pg, _ = _load_pygame()
        pg.display.quit()
        pg.quit()

    def draw(self):
        pg, _ = _load_pygame()
        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.close()
                return
        pg.display.flip()

    def clear(self):
        self.screen.fill(self._pg_color(self.background_color))

    @staticmethod
    def _pg_color(c: Color) -> Any:
        pg, _ = _load_pygame()
        r, g, b = c.to_rgb255()
        return pg.Color(r, g, b)

    def _to_px(self, point: Float[ArrayLike, "2"]) -> tuple[int, int]:
        return self.transform.world_to_px(
            jnp.asarray(point)
        ).tolist()  # pyright: ignore

    def _scale_x(self, length: Float[ArrayLike, ""]) -> int:
        return int(self.transform.scale_length(length))  # pyright: ignore

    def draw_circle(
        self, center: Float[ArrayLike, "2"], radius: Float[ArrayLike, ""], color: Color
    ):
        _, _gfx = _load_pygame()
        _gfx.aacircle(
            self.screen,
            *self._to_px(center),
            self._scale_x(radius),
            self._pg_color(color),
        )
        _gfx.filled_circle(
            self.screen,
            *self._to_px(center),
            self._scale_x(radius),
            self._pg_color(color),
        )

    def draw_line(
        self,
        start: Float[ArrayLike, "2"],
        end: Float[ArrayLike, "2"],
        color: Color,
        width: Float[ArrayLike, ""] = 1,
    ):
        start = jnp.asarray(start)
        end = jnp.asarray(end)

        # Uses rectangle to work around lack of width in aaline
        norm = jnp.array([start[1] - end[1], end[0] - start[0]])  # pyright: ignore
        norm = norm / jnp.linalg.norm(norm) * (width / 2)

        p1 = start + norm
        p2 = start - norm
        p3 = end - norm
        p4 = end + norm

        self.draw_polygon(jnp.asarray([p1, p2, p3, p4]), color)

    def draw_rect(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        _, _gfx = _load_pygame()
        pg = pygame

        top_left = self._to_px(
            jnp.asarray(center) - jnp.array([w / 2, -h / 2])  # pyright: ignore
        )
        width_height = (self._scale_x(w), self._scale_x(h))

        rect = pg.Rect(top_left, width_height)

        _gfx.box(self.screen, rect, self._pg_color(color))

    def draw_polygon(self, points: Float[ArrayLike, "num 2"], color: Color):
        _, _gfx = _load_pygame()
        pts = [self._to_px(point) for point in points]  # pyright: ignore
        if len(pts) >= 3:
            _gfx.aapolygon(self.screen, pts, self._pg_color(color))
            _gfx.filled_polygon(self.screen, pts, self._pg_color(color))
        else:
            raise ValueError("Need at least 3 points to draw a polygon.")

    def draw_text(
        self,
        center: Float[ArrayLike, "2"],
        text: str,
        color: Color,
        size: Float[ArrayLike, ""] = 12,
    ):
        pg, _ = _load_pygame()
        if not pg.font.get_init():
            pg.font.init()
        font = pg.font.SysFont(None, int(size))  # pyright: ignore
        surf = font.render(text, True, self._pg_color(color))
        px, py = self._to_px(center)
        self.screen.blit(surf, (px, py))

    def draw_polyline(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ):
        pg, _ = _load_pygame()
        if len(points) >= 2:  # pyright: ignore
            pg.draw.aalines(
                self.screen,
                self._pg_color(color),
                False,
                [self._to_px(p) for p in points],  # pyright: ignore
            )
        else:
            raise ValueError("Need at least 2 points to draw a polyline.")

    def draw_ellipse(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        _, _gfx = _load_pygame()
        px, py = self._to_px(center)
        rx = max(1, int(self._scale_x(w) / 2))
        ry = max(1, int(self._scale_x(h) / 2))

        _gfx.aaellipse(self.screen, px, py, rx, ry, color=self._pg_color(color))

    def as_array(self) -> Float[ArrayLike, "height width 3"]:
        pg, _ = _load_pygame()
        arr = pg.surfarray.array3d(self.screen).transpose(1, 0, 2).copy()
        return arr



from __future__ import annotations

from pathlib import Path
from typing import ClassVar

import mujoco
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key
from mujoco import mjx

from lerax.space import Box

from .base_mujoco import AbstractMujocoEnv, MujocoEnvState


class Humanoid(AbstractMujocoEnv[Float[Array, "..."], Float[Array, "..."]]):
    """
    MJX-based humanoid environment roughly matching Gymnasium's Humanoid-v5.
    """

    name: ClassVar[str] = "Humanoid"

    action_space: Box
    observation_space: Box
    model: mjx.Model
    mujoco_model: mujoco.MjModel

    frame_skip: int
    dt: Float[Array, ""]

    forward_reward_weight: Float[Array, ""]
    ctrl_cost_weight: Float[Array, ""]
    contact_cost_weight: Float[Array, ""]
    contact_cost_range: Float[Array, "2"]
    healthy_reward: Float[Array, ""]
    terminate_when_unhealthy: bool
    healthy_z_range: Float[Array, "2"]

    reset_noise_scale: Float[Array, ""]
    exclude_current_positions_from_observation: bool
    include_cinert_in_observation: bool
    include_cvel_in_observation: bool
    include_qfrc_actuator_in_observation: bool
    include_cfrc_ext_in_observation: bool

    init_qpos: Float[Array, "..."]
    init_qvel: Float[Array, "..."]

    qpos_size: int
    qvel_size: int
    cinert_size: int
    cvel_size: int
    qfrc_actuator_size: int
    cfrc_ext_size: int

    def __init__(
        self,
        *,
        xml_file: str | Path = "humanoid.xml",
        frame_skip: int = 5,
        forward_reward_weight: float = 1.25,
        ctrl_cost_weight: float = 0.1,
        contact_cost_weight: float = 5e-7,
        contact_cost_range: tuple[float, float] = (-jnp.inf, 10.0),
        healthy_reward: float = 5.0,
        terminate_when_unhealthy: bool = True,
        healthy_z_range: tuple[float, float] = (1.0, 2.0),
        reset_noise_scale: float = 1e-2,
        exclude_current_positions_from_observation: bool = True,
        include_cinert_in_observation: bool = True,
        include_cvel_in_observation: bool = True,
        include_qfrc_actuator_in_observation: bool = True,
        include_cfrc_ext_in_observation: bool = True,
    ):
        asset_path = Path(__file__).resolve().parent / "assets" / xml_file
        if not asset_path.exists():
            raise FileNotFoundError(f"Humanoid asset not found: {asset_path}")

        mj_model = mujoco.MjModel.from_xml_path(str(asset_path))
        mj_data = mujoco.MjData(mj_model)

        self.model = mjx.put_model(mj_model)
        self.mujoco_model = mj_model

        self.frame_skip = int(frame_skip)
        self.dt = jnp.array(mj_model.opt.timestep * self.frame_skip)

        self.init_qpos = jnp.asarray(mj_data.qpos).reshape(-1)
        self.init_qvel = jnp.asarray(mj_data.qvel).reshape(-1)

        self.forward_reward_weight = jnp.array(forward_reward_weight)
        self.ctrl_cost_weight = jnp.array(ctrl_cost_weight)
        self.contact_cost_weight = jnp.array(contact_cost_weight)
        self.contact_cost_range = jnp.asarray(contact_cost_range)
        self.healthy_reward = jnp.array(healthy_reward)
        self.terminate_when_unhealthy = bool(terminate_when_unhealthy)
        self.healthy_z_range = jnp.asarray(healthy_z_range)

        self.reset_noise_scale = jnp.array(reset_noise_scale)
        self.exclude_current_positions_from_observation = bool(
            exclude_current_positions_from_observation
        )
        self.include_cinert_in_observation = bool(include_cinert_in_observation)
        self.include_cvel_in_observation = bool(include_cvel_in_observation)
        self.include_qfrc_actuator_in_observation = bool(
            include_qfrc_actuator_in_observation
        )
        self.include_cfrc_ext_in_observation = bool(include_cfrc_ext_in_observation)

        qpos_size = mj_data.qpos.size
        qvel_size = mj_data.qvel.size
        cinert_size = mj_data.cinert[1:].size
        cvel_size = mj_data.cvel[1:].size
        qfrc_actuator_size = mj_data.qvel.size - 6
        cfrc_ext_size = mj_data.cfrc_ext[1:].size

        skipped_qpos = 2 if self.exclude_current_positions_from_observation else 0

        obs_size = qpos_size + qvel_size
        obs_size -= skipped_qpos
        obs_size += cinert_size if self.include_cinert_in_observation else 0
        obs_size += cvel_size if self.include_cvel_in_observation else 0
        obs_size += (
            qfrc_actuator_size if self.include_qfrc_actuator_in_observation else 0
        )
        obs_size += cfrc_ext_size if self.include_cfrc_ext_in_observation else 0

        self.qpos_size = qpos_size
        self.qvel_size = qvel_size
        self.cinert_size = cinert_size
        self.cvel_size = cvel_size
        self.qfrc_actuator_size = qfrc_actuator_size
        self.cfrc_ext_size = cfrc_ext_size

        bounds = jnp.array(mj_model.actuator_ctrlrange, dtype=jnp.float32)
        low, high = bounds[:, 0], bounds[:, 1]
        self.action_space = Box(low=low, high=high)

        high_obs = jnp.full((obs_size,), jnp.inf, dtype=jnp.float32)
        self.observation_space = Box(low=-high_obs, high=high_obs)

    def initial(self, *, key: Key) -> MujocoEnvState:
        qpos_key, qvel_key = jr.split(key)

        noise_low = -self.reset_noise_scale
        noise_high = self.reset_noise_scale

        qpos_noise = jr.uniform(
            qpos_key, shape=self.init_qpos.shape, minval=noise_low, maxval=noise_high
        )
        qvel_noise = jr.uniform(
            qvel_key, shape=self.init_qvel.shape, minval=noise_low, maxval=noise_high
        )

        qpos = self.init_qpos + qpos_noise
        qvel = self.init_qvel + qvel_noise

        data = mjx.make_data(self.model)
        data = data.replace(qpos=qpos, qvel=qvel)

        return MujocoEnvState(sim_state=data, t=jnp.array(0.0))

    def observation(self, state: MujocoEnvState, *, key: Key) -> Float[Array, "..."]:
        data = state.sim_state

        position = data.qpos.reshape(-1)
        velocity = data.qvel.reshape(-1)

        if self.exclude_current_positions_from_observation:
            position = position[2:]

        if self.include_cinert_in_observation:
            com_inertia = data.cinert[1:].reshape(-1)
        else:
            com_inertia = jnp.zeros((0,), dtype=position.dtype)

        if self.include_cvel_in_observation:
            com_velocity = data.cvel[1:].reshape(-1)
        else:
            com_velocity = jnp.zeros((0,), dtype=position.dtype)

        if self.include_qfrc_actuator_in_observation:
            actuator_forces = data.qfrc_actuator[6:].reshape(-1)
        else:
            actuator_forces = jnp.zeros((0,), dtype=position.dtype)

        if self.include_cfrc_ext_in_observation:
            external_contact_forces = data.cfrc_ext[1:].reshape(-1)
        else:
            external_contact_forces = jnp.zeros((0,), dtype=position.dtype)

        return jnp.concatenate(
            (
                position,
                velocity,
                com_inertia,
                com_velocity,
                actuator_forces,
                external_contact_forces,
            )
        )

    def reward(
        self,
        state: MujocoEnvState,
        action: Float[Array, "..."],
        next_state: MujocoEnvState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        xy_before = self.mass_center(state.sim_state)
        xy_after = self.mass_center(next_state.sim_state)
        xy_velocity = (xy_after - xy_before) / self.dt
        x_velocity = xy_velocity[0]

        healthy_reward = (
            self.is_healthy(next_state.sim_state).astype(float) * self.healthy_reward
        )
        forward_reward = self.forward_reward_weight * x_velocity

        control_cost = self.ctrl_cost_weight * jnp.sum(jnp.square(action))
        contact_cost = self.contact_cost(next_state.sim_state)

        reward = forward_reward + healthy_reward - control_cost - contact_cost
        return reward

    def terminal(self, state: MujocoEnvState, *, key: Key) -> Bool[Array, ""]:
        if not self.terminate_when_unhealthy:
            return jnp.array(False)
        return ~self.is_healthy(state.sim_state)

    def state_info(self, state: MujocoEnvState) -> dict:
        data = state.sim_state
        return {
            "x_position": data.qpos[0],
            "y_position": data.qpos[1],
            "tendon_length": data.ten_length,
            "tendon_velocity": data.ten_velocity,
            "distance_from_origin": jnp.linalg.norm(data.qpos[0:2], ord=2),
        }

    def transition_info(
        self,
        state: MujocoEnvState,
        action: Float[Array, "..."],
        next_state: MujocoEnvState,
    ) -> dict:
        data = next_state.sim_state

        xy_before = self.mass_center(state.sim_state)
        xy_after = self.mass_center(next_state.sim_state)
        xy_velocity = (xy_after - xy_before) / self.dt
        x_velocity, y_velocity = xy_velocity[0], xy_velocity[1]

        healthy_reward = (
            self.is_healthy(next_state.sim_state).astype(float) * self.healthy_reward
        )
        forward_reward = self.forward_reward_weight * x_velocity
        ctrl_cost = self.ctrl_cost_weight * jnp.sum(jnp.square(action))
        contact_cost = self.contact_cost(data)

        return {
            "x_position": data.qpos[0],
            "y_position": data.qpos[1],
            "tendon_length": data.ten_length,
            "tendon_velocity": data.ten_velocity,
            "distance_from_origin": jnp.linalg.norm(data.qpos[0:2], ord=2),
            "x_velocity": x_velocity,
            "y_velocity": y_velocity,
            "reward_survive": healthy_reward,
            "reward_forward": forward_reward,
            "reward_ctrl": -ctrl_cost,
            "reward_contact": -contact_cost,
        }

    def mass_center(self, data: mjx.Data) -> Float[Array, "2"]:
        num = jnp.einsum("b,bj->j", self.model.body_mass, data.xipos)
        denom = jnp.sum(self.model.body_mass)
        com = num / denom
        return com[0:2]

    def is_healthy(self, data: mjx.Data) -> Bool[Array, ""]:
        min_z, max_z = self.healthy_z_range
        z = data.qpos[2]
        return (z > min_z) & (z < max_z)

    def contact_cost(self, data: mjx.Data) -> Float[Array, ""]:
        contact_forces = data.cfrc_ext
        raw_cost = jnp.sum(jnp.square(contact_forces))
        min_cost, max_cost = self.contact_cost_range
        clipped = jnp.clip(raw_cost, min_cost, max_cost)
        return self.contact_cost_weight * clipped



from __future__ import annotations

from typing import ClassVar

import diffrax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from lerax.render import (
    Abstract2DRenderer,
    AbstractRenderer,
    Color,
    PygameRenderer,
    Transform,
)
from lerax.space import Box

from .base_classic_control import (
    AbstractClassicControlEnv,
    AbstractClassicControlEnvState,
)


class ContinuousMountainCarState(AbstractClassicControlEnvState):
    y: Float[Array, "2"]
    t: Float[Array, ""]


class ContinuousMountainCar(
    AbstractClassicControlEnv[
        ContinuousMountainCarState, Float[Array, ""], Float[Array, "2"]
    ]
):
    """
    Continuous Mountain Car environment matching the [Gymnasium Continuous MountainCar environment](https://gymnasium.farama.org/environments/classic_control/continuous_mountain_car/).

    Note:
        To achieve identical dynamics to Gymnasium set `solver=diffrax.Euler()`.

    ## Action Space

    The action space is a 1-dimensional continuous space representing the force applied to the car in the range [-1.0, 1.0].

    ## Observation Space

    The observation space is a 2-dimensional continuous space representing the position and velocity of the car:

    | Index | Observation  | Min Value | Max Value |
    |-------|--------------|-----------|-----------|
    | 0     | Car Position | -1.2      | 0.6       |
    | 1     | Car Velocity | -0.07     | 0.07      |

    Args:
        min_action: Minimum action value (default: -1.0).
        max_action: Maximum action value (default: 1.0).
        min_position: Minimum position of the car (default: -1.2).
        max_position: Maximum position of the car (default: 0.6).
        max_speed: Maximum speed of the car (default: 0.07).
        goal_position: Position at which the goal is reached (default: 0.5).
        power: Power of the car's engine (default: 0.0015).
        dt: Time step for each action (default: 1.0).
        solver: Diffrax solver to use for ODE integration (default: Tsit5).
        stepsize_controller: Step size controller for adaptive solvers (default: PIDController with rtol=1e-5, atol=1e-5).
    """

    name: ClassVar[str] = "ContinuousMountainCar"

    action_space: Box
    observation_space: Box

    min_action: Float[Array, ""]
    max_action: Float[Array, ""]
    min_position: Float[Array, ""]
    max_position: Float[Array, ""]
    max_speed: Float[Array, ""]
    goal_position: Float[Array, ""]
    goal_velocity: Float[Array, ""]

    power: Float[Array, ""]
    low: Float[Array, "2"]
    high: Float[Array, "2"]

    dt: Float[Array, ""]
    solver: diffrax.AbstractSolver
    dt0: Float[Array, ""] | None
    stepsize_controller: diffrax.AbstractStepSizeController

    def __init__(
        self,
        *,
        min_action: Float[ArrayLike, ""] = -1.0,
        max_action: Float[ArrayLike, ""] = 1.0,
        min_position: Float[ArrayLike, ""] = -1.2,
        max_position: Float[ArrayLike, ""] = 0.6,
        max_speed: Float[ArrayLike, ""] = 0.07,
        goal_position: Float[ArrayLike, ""] = 0.5,
        goal_velocity: Float[ArrayLike, ""] = 0.0,
        power: Float[ArrayLike, ""] = 0.0015,
        dt: Float[ArrayLike, ""] = 1.0,
        solver: diffrax.AbstractSolver | None = None,
        stepsize_controller: diffrax.AbstractStepSizeController | None = None,
    ):
        self.min_action = jnp.array(min_action)
        self.max_action = jnp.array(max_action)
        self.min_position = jnp.array(min_position)
        self.max_position = jnp.array(max_position)
        self.max_speed = jnp.array(max_speed)
        self.goal_position = jnp.array(goal_position)
        self.goal_velocity = jnp.array(goal_velocity)
        self.power = jnp.array(power)

        self.low = jnp.array([self.min_position, -self.max_speed])
        self.high = jnp.array([self.max_position, self.max_speed])

        self.action_space = Box(self.min_action, self.max_action)
        self.observation_space = Box(self.low, self.high)

        self.dt = jnp.array(dt)
        self.solver = solver or diffrax.Tsit5()
        is_adaptive = isinstance(self.solver, diffrax.AbstractAdaptiveSolver)
        self.dt0 = None if is_adaptive else self.dt
        if stepsize_controller is None:
            self.stepsize_controller = (
                diffrax.PIDController(rtol=1e-5, atol=1e-5)
                if is_adaptive
                else diffrax.ConstantStepSize()
            )
        else:
            self.stepsize_controller = stepsize_controller

    def initial(self, *, key: Key) -> ContinuousMountainCarState:
        return ContinuousMountainCarState(
            y=jnp.asarray([jr.uniform(key, minval=-0.6, maxval=-0.4), 0.0]),
            t=jnp.array(0.0),
        )

    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, "2"], action: Float[Array, ""]
    ) -> Float[Array, "2"]:
        x, x_d = y
        a = jnp.clip(action, self.min_action, self.max_action)
        x_dd = self.power * a - 0.0025 * jnp.cos(3.0 * x)
        return jnp.array([x_d, x_dd])

    def clip(self, y: Float[Array, "2"]) -> Float[Array, "2"]:
        x, v = y
        v = jnp.clip(v, -self.max_speed, self.max_speed)
        x = jnp.clip(x, self.min_position, self.max_position)
        return jnp.array([x, v])

    def observation(
        self, state: ContinuousMountainCarState, *, key: Key
    ) -> Float[Array, "2"]:
        return state.y

    def reward(
        self,
        state: ContinuousMountainCarState,
        action: Float[Array, ""],
        next_state: ContinuousMountainCarState,
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return (
            100.0 * self.terminal(state, key=key).astype(float)
            - 0.1 * jnp.clip(action, self.min_action, self.max_action) ** 2
        )

    def terminal(
        self, state: ContinuousMountainCarState, *, key: Key
    ) -> Bool[Array, ""]:
        x, v = state.y
        return (x >= self.goal_position) & (v >= self.goal_velocity)

    def render(self, state: ContinuousMountainCarState, renderer: AbstractRenderer):
        if not isinstance(renderer, Abstract2DRenderer):
            raise ValueError(
                "ContinuousMountainCar environment requires a 2D renderer."
            )

        x = state.y[0]

        renderer.clear()

        # Track
        xs = jnp.linspace(self.min_position - 0.5, self.max_position + 0.1, 64)
        ys = jnp.sin(3 * xs) * 0.45
        track_points = jnp.stack([xs, ys], axis=1)
        track_color = Color(0.0, 0.0, 0.0)
        renderer.draw_polyline(track_points, color=track_color)

        # Flag
        flag_h = 0.2
        flag_start = jnp.array(
            [self.goal_position, jnp.sin(3 * self.goal_position) * 0.45]
        )
        flag_end = flag_start + jnp.array([0.0, flag_h])

        flag_pole_color = Color(0.0, 0.0, 0.0)
        flag_color = Color(0.86, 0.24, 0.24)
        renderer.draw_line(flag_start, flag_end, color=flag_pole_color, width=0.005)
        flag_points = jnp.array(
            [
                flag_end,
                flag_end + jnp.array([0.1, -0.03]),
                flag_end + jnp.array([0.0, -0.06]),
            ]
        )
        renderer.draw_polygon(flag_points, color=flag_color)

        # Car
        car_h, car_w, wheel_r, clearance = 0.04, 0.1, 0.02, 0.025
        angle = jnp.arctan2(jnp.cos(3 * x) * 0.45 * 3, 1.0)
        rot = jnp.array(
            [[jnp.cos(angle), -jnp.sin(angle)], [jnp.sin(angle), jnp.cos(angle)]]
        )
        ## Body
        car_col = Color(0.0, 0.0, 0.0)
        clearance_vec = rot @ jnp.array([0.0, clearance + car_h / 2])
        car_center = jnp.array([x, jnp.sin(3 * x) * 0.45]) + clearance_vec

        car_corners = jnp.array(
            [
                [-car_w / 2, -car_h / 2],
                [car_w / 2, -car_h / 2],
                [car_w / 2, car_h / 2],
                [-car_w / 2, car_h / 2],
            ]
        )
        car_corners = (rot @ car_corners.T).T + car_center
        renderer.draw_polygon(car_corners, color=car_col)
        ## Wheels
        wheel_col = Color(0.3, 0.3, 0.3)
        wheel_clearance = rot @ jnp.array([0.0, wheel_r])
        wheel_centers = jnp.array(
            [
                [-car_w / 4, 0.0],
                [car_w / 4, 0.0],
            ]
        )
        wheel_centers = (
            (rot @ wheel_centers.T).T
            + jnp.array([x, jnp.sin(3 * x) * 0.45])
            + wheel_clearance
        )
        renderer.draw_circle(wheel_centers[0], radius=wheel_r, color=wheel_col)
        renderer.draw_circle(wheel_centers[1], radius=wheel_r, color=wheel_col)

        renderer.draw()

    def default_renderer(self) -> Abstract2DRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=350.0,
            offset=jnp.array([width * 0.7, height * 0.4]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )



from .base_renderer import (
    BLACK,
    BLUE,
    GRAY,
    GREEN,
    RED,
    WHITE,
    Abstract2DRenderer,
    Abstract3DRenderer,
    AbstractRenderer,
    Color,
    Transform,
)
from .pygame_renderer import PygameRenderer
from .video import VideoRenderer

__all__ = [
    "AbstractRenderer",
    "Abstract2DRenderer",
    "Abstract3DRenderer",
    "Transform",
    "Color",
    "WHITE",
    "BLACK",
    "GRAY",
    "RED",
    "GREEN",
    "BLUE",
    "PygameRenderer",
    "VideoRenderer",
]



from __future__ import annotations

from typing import Sequence

import jax
from distreqx import distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Integer, Key

from .base_distribution import AbstractMaskableDistribution


class MultiCategorical(AbstractMaskableDistribution[Integer[Array, " dims"]]):
    """
    Product of independent Categorical distributions.

    Attributes:
        distribution: Tuple of underlying Categorical distributions.
        action_dims: Tuple of number of classes for each categorical dimension.

    """

    distribution: tuple[distributions.Categorical, ...]
    action_dims: tuple[int, ...]

    def __init__(
        self,
        logits: (
            Float[ArrayLike, " sum_of_classes"]
            | Sequence[Float[ArrayLike, " classes"]]
            | None
        ) = None,
        probs: (
            Float[ArrayLike, " sum_of_classes"]
            | Sequence[Float[ArrayLike, " classes"]]
            | None
        ) = None,
        action_dims: Sequence[int] | None = None,
    ):
        if (logits is None) == (probs is None):
            raise ValueError("Exactly one of logits or probs must be provided.")

        if action_dims is not None:
            action_dims_tup = tuple(int(d) for d in action_dims)
            if len(action_dims_tup) == 0:
                raise ValueError("action_dims must be non-empty.")
            if any(d <= 0 for d in action_dims_tup):
                raise ValueError("All entries in action_dims must be positive.")
        else:
            action_dims_tup = None

        if logits is not None:
            pieces, inferred_dims = self._split_or_unpack_params(
                logits, action_dims_tup
            )
            self.action_dims = inferred_dims
            self.distribution = tuple(
                distributions.Categorical(logits=piece) for piece in pieces
            )
        else:
            assert probs is not None
            pieces, inferred_dims = self._split_or_unpack_params(probs, action_dims_tup)
            self.action_dims = inferred_dims
            self.distribution = tuple(
                distributions.Categorical(probs=piece) for piece in pieces
            )

    @staticmethod
    def _split_or_unpack_params(
        params: (
            Float[ArrayLike, " sum_of_classes"] | Sequence[Float[ArrayLike, " classes"]]
        ),
        action_dims: tuple[int, ...] | None,
    ) -> tuple[tuple[Float[Array, "... classes"], ...], tuple[int, ...]]:
        # Sequence form: [ ..., classes_i ] for each action dim
        if isinstance(params, (list, tuple)):
            if len(params) == 0:
                raise ValueError("Sequence params must be non-empty.")
            arrays = tuple(jnp.asarray(p) for p in params)
            inferred_dims = tuple(int(a.shape[-1]) for a in arrays)

            if any(d <= 0 for d in inferred_dims):
                raise ValueError(
                    "Each categorical dimension must have at least 1 class."
                )

            if action_dims is not None and inferred_dims != action_dims:
                raise ValueError(
                    f"action_dims={action_dims} does not match params last-dims={inferred_dims}."
                )

            lead_shape = arrays[0].shape[:-1]
            for i, a in enumerate(arrays):
                if a.shape[:-1] != lead_shape:
                    raise ValueError(
                        "All parameter pieces must have the same leading (batch) shape. "
                        f"Piece 0 has {lead_shape}, piece {i} has {a.shape[:-1]}."
                    )

            return arrays, inferred_dims

        # Flat form: [..., sum_of_classes] + action_dims required
        arr = jnp.asarray(params)
        if action_dims is None:
            raise ValueError(
                "action_dims must be provided when logits/probs is a single concatenated array."
            )

        total = int(sum(action_dims))
        if arr.shape[-1] != total:
            raise ValueError(
                f"Last dimension ({arr.shape[-1]}) must equal sum(action_dims) ({total})."
            )

        split_idx = jnp.cumsum(jnp.asarray(action_dims[:-1]))
        pieces = tuple(jnp.split(arr, split_idx, axis=-1))
        return pieces, action_dims

    @property
    def logits(self) -> Float[Array, "... sum_of_classes"]:
        return jnp.concatenate(tuple(d.logits for d in self.distribution), axis=-1)

    @property
    def probs(self) -> Float[Array, "... sum_of_classes"]:
        return jnp.concatenate(tuple(d.probs for d in self.distribution), axis=-1)

    def mask(
        self,
        mask: Bool[Array, "... sum_of_classes"] | Sequence[Bool[Array, "... classes"]],
    ) -> MultiCategorical:
        mask_pieces, _ = self._split_or_unpack_params(mask, self.action_dims)
        masked_logits_pieces = tuple(
            jnp.where(m, d.logits, -jnp.inf)
            for d, m in zip(self.distribution, mask_pieces)
        )
        masked_logits = jnp.concatenate(masked_logits_pieces, axis=-1)
        return MultiCategorical(logits=masked_logits, action_dims=self.action_dims)

    def log_prob(self, value: Integer[ArrayLike, " dims"]) -> Float[Array, "..."]:
        value_arr = jnp.asarray(value)
        if value_arr.shape[-1] != len(self.action_dims):
            raise ValueError(
                f"value last dimension ({value_arr.shape[-1]}) must equal number of action dims "
                f"({len(self.action_dims)})."
            )
        logps = tuple(
            d.log_prob(value_arr[..., i]) for i, d in enumerate(self.distribution)
        )
        return jnp.sum(jnp.stack(logps, axis=-1), axis=-1)

    def prob(self, value: Integer[ArrayLike, " dims"]) -> Float[Array, "..."]:
        return jnp.exp(self.log_prob(value))

    def sample(self, key: Key) -> Integer[Array, " ... dims"]:
        keys = jax.random.split(key, len(self.action_dims))
        samples = tuple(d.sample(k) for d, k in zip(self.distribution, keys))
        return jnp.stack(samples, axis=-1)

    def entropy(self) -> Float[Array, "..."]:
        ents = tuple(d.entropy() for d in self.distribution)
        return jnp.sum(jnp.stack(ents, axis=-1), axis=-1)

    def mean(self) -> Float[Array, " ... dims"]:
        means = tuple(d.mean() for d in self.distribution)
        return jnp.stack(means, axis=-1)

    def mode(self) -> Integer[Array, " ... dims"]:
        modes = tuple(d.mode() for d in self.distribution)
        return jnp.stack(modes, axis=-1)

    def sample_and_log_prob(
        self, key: Key
    ) -> tuple[Integer[Array, " ... dims"], Float[Array, "..."]]:
        keys = jax.random.split(key, len(self.action_dims))
        pairs = tuple(d.sample_and_log_prob(k) for d, k in zip(self.distribution, keys))
        samples = jnp.stack(tuple(p[0] for p in pairs), axis=-1)
        logps = jnp.sum(jnp.stack(tuple(p[1] for p in pairs), axis=-1), axis=-1)
        return samples, logps



from __future__ import annotations

import equinox as eqx
import jax
import optax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Key, Scalar

from lerax.buffer import ReplayBuffer
from lerax.policy import AbstractStatefulDQNPolicy
from lerax.utils import filter_scan

from .off_policy import AbstractOffPolicyAlgorithm, OffPolicyState, OffPolicyStepState


class DQNStats(eqx.Module):
    """
    DQN training statistics.

    Attributes:
        q_loss: The Q-value loss.
        td_error: The temporal difference error.
        mean_q: The mean Q-value.
    """

    q_loss: Float[Array, ""]
    td_error: Float[Array, ""]
    mean_q: Float[Array, ""]


class DQN[PolicyType: AbstractStatefulDQNPolicy](
    AbstractOffPolicyAlgorithm[PolicyType]
):
    """
    Deep Q-Network (DQN) algorithm.

    Attributes:
        optimizer: The optimizer used for training.
        gamma: Discount factor for future rewards.
        tau: Soft update coefficient for target network.
        learning_starts: Number of steps before learning starts.
        target_update_period: Number of updates between target network updates.
        max_grad_norm: Maximum gradient norm for clipping.
        num_envs: Number of parallel environments.
        num_steps: Number of steps to collect per environment.
        gradient_steps: Number of gradient steps per training iteration.
        batch_size: Batch size for training.
        buffer_size: Size of the replay buffer.

    Args:
        gamma: Discount factor for future rewards.
        tau: Soft update coefficient for target network.
        learning_starts: Number of steps before learning starts.
        target_update_period: Number of updates between target network updates.
        max_grad_norm: Maximum gradient norm for clipping.
        num_envs: Number of parallel environments.
        num_steps: Number of steps to collect per environment.
        gradient_steps: Number of gradient steps per training iteration.
        batch_size: Batch size for training.
        buffer_size: Size of the replay buffer.
        learning_rate: Learning rate for the optimizer.
    """

    optimizer: optax.GradientTransformation

    gamma: float
    tau: float
    learning_starts: int
    target_update_period: int
    max_grad_norm: float

    num_envs: int
    num_steps: int
    gradient_steps: int
    batch_size: int
    buffer_size: int

    def __init__(
        self,
        *,
        gamma: float = 0.99,
        tau: float = 1.0,
        learning_starts: int = 128,
        target_update_period: int = 1000,
        max_grad_norm: float = 10.0,
        num_envs: int = 1,
        num_steps: int = 1,
        gradient_steps: int = 1,
        batch_size: int = 32,
        buffer_size: int = 2**20,
        learning_rate: optax.ScalarOrSchedule = 3e-4,
    ):
        self.gamma = gamma
        self.tau = tau
        self.learning_starts = learning_starts
        self.target_update_period = target_update_period
        self.max_grad_norm = max_grad_norm

        self.num_envs = num_envs
        self.num_steps = num_steps
        self.gradient_steps = gradient_steps
        self.batch_size = batch_size
        self.buffer_size = buffer_size

        adam = optax.inject_hyperparams(optax.adam)(learning_rate)
        clip = optax.clip_by_global_norm(self.max_grad_norm)
        self.optimizer = optax.chain(clip, adam)

    def per_step(
        self, step_state: OffPolicyStepState[PolicyType]
    ) -> OffPolicyStepState[PolicyType]:
        return step_state

    def per_iteration(
        self, state: OffPolicyState[PolicyType]
    ) -> OffPolicyState[PolicyType]:
        if self.target_update_period <= 0:
            return state

        updates = state.iteration_count * self.gradient_steps

        policy = state.policy
        current_dynamic = eqx.filter(policy.q_network, eqx.is_inexact_array)
        target_dynamic, target_static = eqx.partition(
            policy.target_q_network, eqx.is_inexact_array
        )

        updated_dynamic = jax.tree.map(
            lambda q, tq: self.tau * q + (1 - self.tau) * tq,
            current_dynamic,
            target_dynamic,
        )

        should_update = jnp.equal(jnp.remainder(updates, self.target_update_period), 0)
        next_target_dynamic = lax.cond(
            should_update, lambda: updated_dynamic, lambda: target_dynamic
        )

        target_q_network = eqx.combine(next_target_dynamic, target_static)
        policy = eqx.tree_at(lambda p: p.target_q_network, policy, target_q_network)

        return eqx.tree_at(lambda s: s.policy, state, policy)

    # Needs to be static so the first argument can be a policy
    # eqx.filter_value_and_grad doesn't support argnums
    @staticmethod
    def dqn_loss(
        policy: PolicyType,
        sample: ReplayBuffer,
        gamma: float,
    ) -> tuple[Float[Array, ""], DQNStats]:
        q_network = policy.q_network
        target_dynamic, target_static = eqx.partition(
            policy.target_q_network, eqx.is_inexact_array
        )
        target_dynamic = lax.stop_gradient(target_dynamic)
        target_q_network = eqx.combine(target_dynamic, target_static)

        target_max = jax.vmap(target_q_network.q_values)(
            sample.next_states, sample.next_observations
        )[1].max(axis=-1)
        td_target = sample.rewards + gamma * target_max * (
            1.0 - sample.dones.astype(float)
        )

        current_q = jnp.take_along_axis(
            jax.vmap(q_network.q_values)(sample.states, sample.observations)[1],
            sample.actions.astype(int)[..., None],
            axis=-1,
        ).squeeze(-1)
        q_loss = optax.huber_loss(current_q, td_target).mean()

        stats = DQNStats(
            q_loss=q_loss,
            td_error=jnp.mean(jnp.abs(td_target - current_q)),
            mean_q=jnp.mean(current_q),
        )

        return q_loss, stats

    dqn_loss_grad = staticmethod(eqx.filter_value_and_grad(dqn_loss, has_aux=True))

    def train_sample(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        sample: ReplayBuffer,
    ) -> tuple[PolicyType, optax.OptState, DQNStats]:
        (_, stats), grads = self.dqn_loss_grad(policy, sample, self.gamma)

        updates, new_opt_state = self.optimizer.update(
            grads, opt_state, eqx.filter(policy, eqx.is_inexact_array)
        )
        policy = eqx.apply_updates(policy, updates)

        return policy, new_opt_state, stats

    def train(
        self,
        policy: PolicyType,
        opt_state: optax.OptState,
        buffer: ReplayBuffer,
        *,
        key: Key,
    ) -> tuple[PolicyType, optax.OptState, dict[str, Scalar]]:
        def sample_scan(
            carry: tuple[PolicyType, optax.OptState], key: Key
        ) -> tuple[tuple[PolicyType, optax.OptState], DQNStats]:
            policy, opt_state = carry
            policy, opt_state, stats = self.train_sample(
                policy, opt_state, buffer.sample(self.batch_size, key=key)
            )
            return (policy, opt_state), stats

        (policy, opt_state), stats = filter_scan(
            sample_scan, (policy, opt_state), jr.split(key, self.gradient_steps)
        )

        stats = jax.tree.map(jnp.mean, stats)
        log = {
            "q_loss": stats.q_loss,
            "td_error": stats.td_error,
            "mean_q": stats.mean_q,
        }
        return policy, opt_state, log



from __future__ import annotations

from abc import abstractmethod
from dataclasses import replace

import diffrax
import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, Bool, Float, Key

from lerax.space import AbstractSpace

from ..base_env import AbstractEnv, AbstractEnvState


class AbstractClassicControlEnvState(AbstractEnvState):
    y: eqx.AbstractVar[Float[Array, "..."]]
    t: eqx.AbstractVar[Float[Array, ""]]


class AbstractClassicControlEnv[
    StateType: AbstractClassicControlEnvState, ActType, ObsType
](AbstractEnv[StateType, ActType, ObsType]):
    name: eqx.AbstractClassVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    dt: eqx.AbstractVar[float]
    solver: eqx.AbstractVar[diffrax.AbstractSolver]
    dt0: eqx.AbstractVar[float | None]
    stepsize_controller: eqx.AbstractVar[diffrax.AbstractStepSizeController]

    @abstractmethod
    def dynamics(
        self, t: Float[Array, ""], y: Float[Array, " n"], action: ActType
    ) -> Float[Array, " n"]:
        """Compute the time derivative of the state."""

    @abstractmethod
    def clip(self, y: Float[Array, " n"]) -> Float[Array, " n"]:
        """Clip the state to be within valid bounds."""

    def transition(self, state: StateType, action: ActType, *, key: Key) -> StateType:
        @diffrax.ODETerm
        def term(t, y, args):
            return self.dynamics(t, y, action)

        saveat = diffrax.SaveAt(t1=True)
        sol = diffrax.diffeqsolve(
            term,
            solver=self.solver,
            t0=state.t,
            t1=state.t + self.dt,
            dt0=self.dt0,
            y0=state.y,
            args=action,
            saveat=saveat,
            stepsize_controller=self.stepsize_controller,
        )
        assert sol.ys is not None

        return replace(state, y=self.clip(sol.ys[0]), t=state.t + self.dt)

    def truncate(self, state: StateType) -> Bool[Array, ""]:
        return jnp.array(False)

    def state_info(self, state: StateType) -> dict:
        return {}

    def transition_info(
        self,
        state: StateType,
        action: ActType,
        next_state: StateType,
    ) -> dict:
        return {}



from __future__ import annotations

import operator
from functools import reduce
from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from .base_space import AbstractSpace
from .utils import try_cast


class Box(AbstractSpace[Float[Array, " ..."]]):
    """
    A space of continuous values.

    Values in the Box space can be unbounded (i.e., infinite) on either
    the lower or upper end. Sampling from an unbounded dimension uses a
    normal distribution. Sampling from a dimension that is only bounded on
    one end uses an exponential distribution shifted to the bound. Sampling
    a bounded dimension uses a uniform distribution.

    Attributes:
        low: The lower bounds of the space.
        high: The upper bounds of the space.
        shape: The shape of the space.

    Args:
        low: The lower bounds of the space.
        high: The upper bounds of the space.
        shape: The shape of the space. If None, the shape is inferred from low
            and high.
    """

    _shape: tuple[int, ...]
    low: Float[Array, "#n"]
    high: Float[Array, "#n"]

    def __init__(
        self,
        low: Float[ArrayLike, "#n"],
        high: Float[ArrayLike, "#n"],
        shape: tuple[int, ...] | None = None,
    ):
        low = jnp.asarray(low, dtype=float)
        high = jnp.asarray(high, dtype=float)
        if shape is None:
            low, high = jnp.broadcast_arrays(low, high)
            shape = low.shape
            # TODO: Add warning if both shapes change

        assert low.shape == high.shape, "Box low and high must have the same shape"

        self._shape = shape
        self.low = jnp.broadcast_to(low, shape)
        self.high = jnp.broadcast_to(high, shape)

    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    def canonical(self) -> Float[Array, " ..."]:
        return (self.low + self.high) / 2

    def sample(self, key: Key) -> Float[Array, " ..."]:
        bounded_key, unbounded_key, upper_bounded_key, lower_bounded_key = jr.split(
            key, 4
        )

        bounded_above = jnp.isfinite(self.high)
        bounded_below = jnp.isfinite(self.low)

        bounded = bounded_above & bounded_below
        unbounded = ~bounded_above & ~bounded_below
        upper_bounded = ~bounded_below & bounded_above
        lower_bounded = bounded_below & ~bounded_above

        sample = jnp.empty(self._shape, dtype=self.low.dtype)

        sample = jnp.where(
            bounded,
            jr.uniform(bounded_key, self._shape, minval=self.low, maxval=self.high),
            sample,
        )

        sample = jnp.where(unbounded, jr.normal(unbounded_key, self._shape), sample)

        sample = jnp.where(
            upper_bounded,
            self.high - jr.exponential(upper_bounded_key, self._shape),
            sample,
        )

        sample = jnp.where(
            lower_bounded,
            self.low + jr.exponential(lower_bounded_key, self._shape),
            sample,
        )

        return sample

    def contains(self, x: Any) -> Bool[Array, ""]:
        x = try_cast(x)
        if x is None:
            return jnp.array(False)

        if x.shape != self._shape:
            return jnp.array(False)

        return jnp.all(x >= self.low) & jnp.all(x <= self.high)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Box):
            return False

        return bool(jnp.array_equal(self.low, other.low)) and bool(
            jnp.array_equal(self.high, other.high)
        )

    def __repr__(self) -> str:
        return f"Box(low={self.low}, high={self.high})"

    def __hash__(self) -> int:
        return hash((self.low.tobytes(), self.high.tobytes()))

    def flatten_sample(self, sample: Float[Array, " ..."]) -> Float[Array, " n"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> int:
        return reduce(operator.mul, self._shape)



from .base_callback import (
    AbstractCallback,
    AbstractCallbackState,
    AbstractCallbackStepState,
    AbstractContinueTrainingCallback,
    AbstractIterationCallback,
    AbstractStatelessCallback,
    AbstractStepCallback,
    AbstractTrainingCallback,
    IterationContext,
    ResetContext,
    StepContext,
    TrainingContext,
)
from .empty import EmptyCallback
from .list import CallbackList
from .progress_bar import ProgressBarCallback
from .tensorboard import TensorBoardCallback

__all__ = [
    "AbstractCallback",
    "AbstractCallbackState",
    "AbstractCallbackStepState",
    "AbstractContinueTrainingCallback",
    "AbstractIterationCallback",
    "AbstractStatelessCallback",
    "AbstractStepCallback",
    "AbstractTrainingCallback",
    "IterationContext",
    "ResetContext",
    "StepContext",
    "TrainingContext",
    "EmptyCallback",
    "CallbackList",
    "ProgressBarCallback",
    "TensorBoardCallback",
]



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import jax
from jax import numpy as jnp
from jaxtyping import Key


class AbstractBuffer(eqx.Module):
    def resolve_axes(
        self,
        batch_axes: tuple[int, ...] | int | None,
    ) -> tuple[int, ...]:
        ndim = len(self.shape)

        if batch_axes is None:
            axes = tuple(range(ndim))
        elif isinstance(batch_axes, int):
            axes = (batch_axes,)
        else:
            axes = tuple(batch_axes)

        axes = tuple(a + ndim if a < 0 else a for a in axes)
        if len(set(axes)) != len(axes) or any(a < 0 or a >= ndim for a in axes):
            raise ValueError(f"Invalid batch_axes {batch_axes} for array ndim={ndim}.")

        return axes

    def flatten_axes[SelfType: AbstractBuffer](
        self: SelfType,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> SelfType:
        axes = self.resolve_axes(batch_axes)
        num_axes = len(axes)
        target_axes = tuple(range(num_axes))
        max_axis = max(axes) if axes else -1

        def flatten_leaf(x):
            if not isinstance(x, jnp.ndarray):
                return x

            if x.ndim <= max_axis:
                return x

            moved = jnp.moveaxis(x, axes, target_axes)

            leading = 1
            for i in range(num_axes):
                leading *= moved.shape[i]

            return moved.reshape((leading,) + moved.shape[num_axes:])

        return jax.tree.map(flatten_leaf, self)

    @abstractmethod
    def batches[SelfType: AbstractBuffer](
        self: SelfType,
        batch_size: int,
        *,
        key: Key | None = None,
        batch_axes: tuple[int, ...] | int | None = None,
    ) -> SelfType:
        """Return an iterator over batches of data from the buffer."""

    @abstractmethod
    def sample[SelfType: AbstractBuffer](
        self: SelfType,
        batch_size: int,
        *,
        key: Key,
    ) -> SelfType:
        """Return uniformly sampled batch of data from the buffer."""

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the buffer."""



from __future__ import annotations

from abc import abstractmethod
from typing import Sequence, cast

import equinox as eqx
import optax
from jax import random as jr
from jaxtyping import Array, Int, Key

from lerax.callback import (
    AbstractCallback,
    AbstractCallbackState,
    AbstractCallbackStepState,
    CallbackList,
    TrainingContext,
)
from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.policy import AbstractPolicy, AbstractPolicyState, AbstractStatefulPolicy
from lerax.utils import filter_scan


class AbstractStepState(eqx.Module):
    """
    Base class for algorithm state that is vectorized over environment steps.

    Attributes:
        env_state: The state of the environment.
        policy_state: The state of the policy.
        callback_state: The state of the callback for this step.
    """

    env_state: eqx.AbstractVar[AbstractEnvLikeState]
    policy_state: eqx.AbstractVar[AbstractPolicyState]
    callback_state: eqx.AbstractVar[AbstractCallbackStepState]

    def with_callback_state[A: AbstractStepState](
        self: A, callback_state: AbstractCallbackStepState | None
    ) -> A:
        """
        Return a new step state with the given callback state.

        Args:
            callback_state: The new callback state. If None, the existing state is used.

        Returns:
            A new step state with the updated callback state.
        """
        return eqx.tree_at(lambda s: s.callback_state, self, callback_state)


class AbstractAlgorithmState[PolicyType: AbstractStatefulPolicy](eqx.Module):
    """
    Base class for algorithm states.

    Attributes:
        iteration_count: The current iteration count.
        step_state: The state for the current step.
        env: The environment being used.
        policy: The policy being trained.
        opt_state: The optimizer state.
        callback_state: The state of the callback for this iteration.
    """

    iteration_count: eqx.AbstractVar[Int[Array, ""]]
    step_state: eqx.AbstractVar[AbstractStepState]
    env: eqx.AbstractVar[AbstractEnvLike]
    policy: eqx.AbstractVar[PolicyType]
    opt_state: eqx.AbstractVar[optax.OptState]
    callback_state: eqx.AbstractVar[AbstractCallbackState]

    def next[A: AbstractAlgorithmState](
        self: A,
        step_state: AbstractStepState,
        policy: PolicyType,
        opt_state: optax.OptState,
    ) -> A:
        """
        Return a new algorithm state for the next iteration.

        Increments the iteration count and updates the step state, policy, and
        optimizer state.

        Args:
            step_state: The new step state.
            policy: The new policy.
            opt_state: The new optimizer state.

        Returns:
            A new algorithm state with the updated fields.
        """
        return eqx.tree_at(
            lambda s: (s.iteration_count, s.step_state, s.policy, s.opt_state),
            self,
            (self.iteration_count + 1, step_state, policy, opt_state),
        )

    def with_callback_states[A: AbstractAlgorithmState](
        self: A, callback_state: AbstractCallbackState
    ) -> A:
        """
        Return a new algorithm state with the given callback state.

        Args:
            callback_state: The new callback state.

        Returns:
            A new algorithm state with the updated callback state.
        """
        return eqx.tree_at(lambda s: s.callback_state, self, callback_state)


class AbstractAlgorithm[
    PolicyType: AbstractStatefulPolicy, StateType: AbstractAlgorithmState
](eqx.Module):
    """
    Base class for RL algorithms.

    Provides the main training loop and abstract methods for algorithm-specific behavior.

    Attributes:
        optimizer: The optimizer used for training.
        num_envs: The number of parallel environments.
        num_steps: The number of steps per environment per iteration.
    """

    optimizer: eqx.AbstractVar[optax.GradientTransformation]

    num_envs: eqx.AbstractVar[int]
    num_steps: eqx.AbstractVar[int]

    @abstractmethod
    def reset(
        self,
        env: AbstractEnvLike,
        policy: PolicyType,
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> StateType:
        """
        Return the initial carry for the training iteration.

        Args:
            env: The environment to train on.
            policy: The policy to train.
            key: A JAX PRNG key.
            callback: A callback or list of callbacks to use during training.

        Returns:
            The initial algorithm state.
        """

    @abstractmethod
    def per_iteration(self, state: StateType) -> StateType:
        """
        Process the algorithm state after each iteration.

        Used for algorithm-specific bookkeeping.

        Args:
            state: The current algorithm state.

        Returns:
            The updated algorithm state.
        """

    @abstractmethod
    def iteration(
        self,
        state: StateType,
        *,
        key: Key,
        callback: AbstractCallback,
    ) -> StateType:
        """
        Perform a single iteration of training.

        Args:
            state: The current algorithm state.
            key: A JAX PRNG key.
            callback: A callback or list of callbacks to use during training.

        Returns:
            The updated algorithm state.
        """

    @abstractmethod
    def num_iterations(self, total_timesteps: int) -> int:
        """Number of iterations per training session."""

    @eqx.filter_jit
    def learn[A: AbstractPolicy](
        self,
        env: AbstractEnvLike,
        policy: A,
        total_timesteps: int,
        *,
        key: Key,
        callback: Sequence[AbstractCallback] | AbstractCallback | None = None,
    ) -> A:
        """
        Train the policy on the environment for a given number of timesteps.

        Args:
            env: The environment to train on.
            policy: The policy to train.
            total_timesteps: The total number of timesteps to train for.
            key: A JAX PRNG key.
            callback: A callback or list of callbacks to use during training.

        Returns:
            policy: The trained policy.
        """
        callback_start_key, reset_key, learn_key, callback_end_key = jr.split(key, 4)

        if callback is None:
            callback = []
        if not isinstance(callback, AbstractCallback):
            callbacks = list(callback)
            callback = CallbackList(callbacks=callbacks)

        if isinstance(policy, AbstractStatefulPolicy):
            _policy = cast(PolicyType, policy)
        else:
            _policy = cast(PolicyType, policy.into_stateful())

        state = self.reset(env, _policy, key=reset_key, callback=callback)

        state = state.with_callback_states(
            callback.on_training_start(
                ctx=TrainingContext(
                    state.callback_state,
                    state.step_state.callback_state,
                    env,
                    state.policy,
                    total_timesteps,
                    state.iteration_count,
                    state.opt_state,
                    locals(),
                ),
                key=callback_start_key,
            )
        )

        state, _ = filter_scan(
            lambda s, k: (self.iteration(s, key=k, callback=callback), None),
            state,
            jr.split(learn_key, self.num_iterations(total_timesteps)),
        )

        state = state.with_callback_states(
            callback.on_training_end(
                ctx=TrainingContext(
                    state.callback_state,
                    state.step_state.callback_state,
                    env,
                    state.policy,
                    total_timesteps,
                    state.iteration_count,
                    state.opt_state,
                    locals(),
                ),
                key=callback_end_key,
            )
        )

        if isinstance(policy, AbstractStatefulPolicy):
            return state.policy
        else:
            return state.policy.into_stateless()



from __future__ import annotations

import jax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.policy import (
    AbstractPolicy,
    AbstractPolicyState,
    AbstractStatefulPolicy,
)


def rollout_while(
    env: AbstractEnvLike,
    policy: AbstractStatefulPolicy,
    *,
    key: Key,
    deterministic: bool = False,
) -> Float[Array, ""]:
    def continue_rollout(
        carry: tuple[AbstractEnvLikeState, AbstractPolicyState, Key, Float[Array, ""]],
    ) -> Bool[Array, ""]:
        env_state, _, key, _ = carry
        return ~(env.terminal(env_state, key=key) | env.truncate(env_state))

    def step(
        carry: tuple[AbstractEnvLikeState, AbstractPolicyState, Key, Float[Array, ""]],
    ):
        env_state, policy_state, key, cumulative_reward = carry
        carry_key, obs_key, action_key, transition_key = jr.split(key, 4)

        obs = env.observation(env_state, key=obs_key)

        if deterministic:
            next_policy_state, action = policy(policy_state, obs)
        else:
            next_policy_state, action = policy(policy_state, obs, key=action_key)

        next_env_state = env.transition(env_state, action, key=transition_key)
        reward = env.reward(env_state, action, next_env_state, key=carry_key)

        return (
            next_env_state,
            next_policy_state,
            carry_key,
            cumulative_reward + reward,
        )

    env_state = env.initial(key=key)
    policy_state = policy.reset(key=key)
    carry = (env_state, policy_state, key, jnp.array(0.0))
    carry = jax.lax.while_loop(continue_rollout, step, carry)
    return carry[3]


def rollout_scan(
    env: AbstractEnvLike,
    policy: AbstractStatefulPolicy,
    *,
    key: Key,
    deterministic: bool = False,
    max_steps: int = 1024,
) -> Float[Array, ""]:

    def step(
        carry: tuple[AbstractEnvLikeState, AbstractPolicyState, Bool[Array, ""]],
        key: Key,
    ) -> tuple[
        tuple[AbstractEnvLikeState, AbstractPolicyState, Bool[Array, ""]],
        Float[Array, ""],
    ]:
        env_state, policy_state, done = carry

        def next_step():
            carry_key, obs_key, action_key, transition_key, terminal_key = jr.split(
                key, 5
            )

            obs = env.observation(env_state, key=obs_key)

            if deterministic:
                next_policy_state, action = policy(policy_state, obs)
            else:
                next_policy_state, action = policy(policy_state, obs, key=action_key)

            next_env_state = env.transition(env_state, action, key=transition_key)
            reward = env.reward(env_state, action, next_env_state, key=carry_key)
            done = env.terminal(next_env_state, key=terminal_key) | env.truncate(
                next_env_state
            )

            return (next_env_state, next_policy_state, done), reward

        def done_step():
            return (env_state, policy_state, jnp.array(True)), jnp.array(0.0)

        return jax.lax.cond(done, done_step, next_step)

    env_state = env.initial(key=key)
    policy_state = policy.reset(key=key)
    carry = (env_state, policy_state, jnp.array(False))
    _, rewards = jax.lax.scan(step, carry, jr.split(key, max_steps))
    return jnp.sum(rewards)


def average_reward(
    env: AbstractEnvLike,
    policy: AbstractPolicy,
    num_episodes: int = 4,
    max_steps: int | None = 1024,
    deterministic: bool = False,
    *,
    key: Key,
) -> Float[Array, ""]:
    if not isinstance(policy, AbstractStatefulPolicy):
        policy = policy.into_stateful()

    def episode_reward(key: Key) -> Float[Array, ""]:
        if max_steps is None:
            return rollout_while(env, policy, key=key, deterministic=deterministic)
        else:
            return rollout_scan(
                env, policy, key=key, deterministic=deterministic, max_steps=max_steps
            )

    rewards = jax.vmap(episode_reward)(jr.split(key, num_episodes))
    return jnp.mean(rewards)



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class Discrete(AbstractSpace[Int[Array, ""]]):
    """
    A space of finite discrete values.

    Attributes:
        n: The number of discrete values.

    Args:
        n: The number of discrete values.
    """

    n: int

    def __init__(self, n: int):
        assert n > 0, "n must be positive"  # pyright: ignore

        self.n = n

    @property
    def shape(self) -> tuple[int, ...]:
        return ()

    def canonical(self) -> Int[Array, ""]:
        return jnp.array(0, dtype=int)

    def sample(self, key: Key) -> Int[Array, ""]:
        return jr.randint(key, shape=(), minval=0, maxval=self.n)

    def contains(self, x: Any) -> Bool[Array, ""]:
        x = try_cast(x)
        if x is None:
            return jnp.array(False)

        if x.ndim != 0:
            return jnp.array(False)
        x = x.squeeze()

        if ~jnp.array_equal(x, jnp.floor(x)):
            return jnp.array(False)

        return 0 <= x < self.n

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Discrete):
            return False
        return self.n == other.n

    def __repr__(self) -> str:
        return f"Discrete({self.n})"

    def __hash__(self) -> int:
        return hash(self.n)

    def flatten_sample(self, sample: Int[Array, ""]) -> Float[Array, " 1"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> int:
        return 1



from __future__ import annotations

from typing import ClassVar

from jaxtyping import Array, Integer, Key, Real

from lerax.env import AbstractEnvLike, AbstractEnvLikeState
from lerax.space import AbstractSpace, Discrete

from ..q import MLPQPolicy
from .base_dqn import AbstractStatelessDQNPolicy


class MLPDQNPolicy[ObsType: Real[Array, "..."]](
    AbstractStatelessDQNPolicy[ObsType, MLPQPolicy]
):
    """
    Deep Q-Network (DQN) policy with MLP Q-network components.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        q_network: The Q-network used for action selection.
        target_q_network: The target Q-network used for stable learning.

    Args:
        env: The environment to create the policy for.
        epsilon: The epsilon value for epsilon-greedy action selection.
        width_size: The width of the hidden layers in the MLP.
        depth: The number of hidden layers in the MLP.
        key: JAX PRNG key for parameter initialization.

    Raises:
        ValueError: If the environment's action space is not Discrete.
    """

    name: ClassVar[str] = "MLPDQNPolicy"

    action_space: Discrete
    observation_space: AbstractSpace[ObsType]

    q_network: MLPQPolicy
    target_q_network: MLPQPolicy

    def __init__[StateType: AbstractEnvLikeState](
        self,
        env: AbstractEnvLike[StateType, Integer[Array, ""], ObsType],
        *,
        epsilon: float = 0.1,
        width_size: int = 64,
        depth: int = 2,
        key: Key,
    ):
        if not isinstance(env.action_space, Discrete):
            raise ValueError(
                f"MLPQPolicy only supports Discrete action spaces, got {type(env.action_space)}"
            )

        self.q_network = MLPQPolicy(
            env, epsilon=epsilon, width_size=width_size, depth=depth, key=key
        )
        self.target_q_network = MLPQPolicy(
            env, epsilon=epsilon, width_size=width_size, depth=depth, key=key
        )
        self.action_space = env.action_space
        self.observation_space = env.observation_space



from __future__ import annotations

import equinox as eqx
from jaxtyping import Array, Integer, Key

from lerax.space import AbstractSpace, Discrete

from ..base_policy import (
    AbstractPolicyState,
    AbstractStatefulPolicy,
    AbstractStatefulWrapper,
    AbstractStatelessPolicy,
    NullStatefulPolicyState,
)
from ..q import AbstractStatefulQPolicy, AbstractStatelessQPolicy, QStatefulWrapper


class AbstractStatelessDQNPolicy[ObsType, PolicyType: AbstractStatelessQPolicy](
    AbstractStatelessPolicy[Integer[Array, ""], ObsType]
):
    """
    Base class for stateless Deep Q-Network (DQN) policies.

    DQN policies utilize two Q-networks: a primary Q-network for action
    selection and a target Q-network for stable learning.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        q_network: The primary Q-network used for action selection.
        target_q_network: The target Q-network used for stable learning.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[Discrete]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    q_network: PolicyType
    target_q_network: PolicyType

    def q_values(self, observation: ObsType) -> Array:
        return self.q_network.q_values(observation)

    def target_q_values(self, observation: ObsType) -> Array:
        return self.target_q_network.q_values(observation)

    def __call__(
        self, observation: ObsType, *, key: Key | None = None
    ) -> Integer[Array, ""]:
        return self.q_network(observation, key=key)

    def into_stateful[SelfType: AbstractStatelessDQNPolicy](
        self: SelfType,
    ) -> DQNStatefulWrapper[SelfType, ObsType]:
        return DQNStatefulWrapper(self)


class AbstractStatefulDQNPolicy[
    StateType: AbstractPolicyState, ObsType, PolicyType: AbstractStatefulQPolicy
](AbstractStatefulPolicy[StateType, Integer[Array, ""], ObsType]):
    """
    Base class for stateful Deep Q-Network (DQN) policies.

    DQN policies utilize two Q-networks: a primary Q-network for action
    selection and a target Q-network for stable learning.

    In a stateful DQN policy the Q-networks should be identical in structure
    and differ only in their parameters. Due to this the state is shared.

    Attributes:
        name: Name of the policy class.
        action_space: The action space of the environment.
        observation_space: The observation space of the environment.
        q_network: The primary Q-network used for action selection.
        target_q_network: The target Q-network used for stable learning.
    """

    name: eqx.AbstractClassVar[str]
    action_space: eqx.AbstractVar[Discrete]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    q_network: PolicyType
    target_q_network: PolicyType

    def __call__(
        self, state: StateType, observation: ObsType, *, key: Key | None = None
    ) -> tuple[StateType, Integer[Array, ""]]:
        """
        Return the next state and action for a given observation and state.

        Acts by selecting the action with the highest Q-value or random action
        based on epsilon-greedy exploration.

        Args:
            state: The current internal state of the policy.
            observation: The current observation from the environment.
            key: JAX PRNG key for stochastic action selection. If None, the
                action with the highest Q-value is always selected.

        Returns:
            Tuple of the next internal state and the selected action.
        """
        return self.q_network(state, observation, key=key)

    def q_values(
        self, state: StateType, observation: ObsType
    ) -> tuple[StateType, Array]:
        """
        Return Q-values for all actions given an observation and state.

        Args:
            state: The current internal state of the policy.
            observation: The current observation from the environment.

        Returns:
            Tuple of the next internal state and the Q-values for all actions.
        """
        return self.q_network.q_values(state, observation)

    def target_q_values(
        self, state: StateType, observation: ObsType
    ) -> tuple[StateType, Array]:
        """
        Return target Q-values for all actions given an observation and state.

        Args:
            state: The current internal state of the policy.
            observation: The current observation from the environment.

        Returns:
            Tuple of the next internal state and the target Q-values for all actions.
        """
        return self.target_q_network.q_values(state, observation)


class DQNStatefulWrapper[PolicyType: AbstractStatelessDQNPolicy, ObsType](
    AbstractStatefulDQNPolicy[
        NullStatefulPolicyState, ObsType, AbstractStatefulQPolicy
    ],
    AbstractStatefulWrapper[PolicyType, Integer[Array, ""], ObsType],
):
    _policy: PolicyType
    q_network: AbstractStatefulQPolicy
    target_q_network: AbstractStatefulQPolicy

    def __init__(self, policy: PolicyType):
        self._policy = policy
        self.q_network = QStatefulWrapper(policy.q_network)
        self.target_q_network = QStatefulWrapper(policy.target_q_network)

    def reset(self, *, key):
        return NullStatefulPolicyState()

    @property
    def policy(self) -> PolicyType:
        return eqx.tree_at(
            lambda p: (p.q_network, p.target_q_network),
            self._policy,
            (self.q_network, self.target_q_network),
        )


type AbstractDQNPolicy = AbstractStatefulDQNPolicy | AbstractStatelessDQNPolicy



from __future__ import annotations

import equinox as eqx
from distreqx import bijectors, distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Float

from .base_distribution import AbstractTransformedDistribution


class SquashedNormal(AbstractTransformedDistribution[Float[Array, " dims"]]):
    """
    Normal distribution with squashing bijector for bounded outputs.

    Attributes:
        distribution: The underlying distreqx Transformed distribution.

    Args:
        loc: The mean of the normal distribution.
        scale: The standard deviation of the normal distribution.
        high: The upper bound for bounded squashing.
        low: The lower bound for bounded squashing.
    """

    distribution: distributions.Transformed

    def __init__(
        self,
        loc: Float[ArrayLike, ""],
        scale: Float[ArrayLike, ""],
        high: Float[ArrayLike, ""] = jnp.array(1.0),
        low: Float[ArrayLike, ""] = jnp.array(-1.0),
    ):
        loc = jnp.asarray(loc)
        scale = jnp.asarray(scale)

        (high, low) = eqx.error_if(
            (high, low),
            ~(jnp.isfinite(high) & jnp.isfinite(low)),
            "SquashedNormal requires finite low/high for all "
            "dimensions. Got non-finite bounds.",
        )

        high = jnp.asarray(high)
        low = jnp.asarray(low)

        normal = distributions.Normal(loc=loc, scale=scale)

        sigmoid = bijectors.Sigmoid()
        affine = bijectors.ScalarAffine(scale=(high - low), shift=low)
        bijector = bijectors.Chain((affine, sigmoid))

        self.distribution = distributions.Transformed(normal, bijector)

    @property
    def loc(self) -> Float[Array, " dims"]:
        assert isinstance(self.distribution._distribution, distributions.Normal)
        return self.distribution._distribution.loc

    @property
    def scale(self) -> Float[Array, " dims"]:
        assert isinstance(self.distribution._distribution, distributions.Normal)
        return self.distribution._distribution.scale



from __future__ import annotations

from abc import abstractmethod
from typing import Any

import equinox as eqx
from jaxtyping import Array, Bool, Float, Key


class AbstractSpace[SampleType](eqx.Module):
    """
    Abstract base class for defining a space.

    A space is a set of values that can be sampled from.
    """

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...] | None:
        """The shape of a sample of the space."""

    @abstractmethod
    def sample(self, key: Key) -> SampleType:
        """Returns a random sample from the space."""

    @abstractmethod
    def contains(self, x: Any) -> Bool[Array, ""]:
        """Returns True if the input is in the space, False otherwise."""

    @abstractmethod
    def flatten_sample(self, sample: SampleType) -> Float[Array, " n"]:
        """Flatten a sample from the space into a 1-D array."""

    @property
    @abstractmethod
    def flat_size(self) -> int:
        """Return the size of a flattened sample from the space."""

    @abstractmethod
    def canonical(self) -> SampleType:
        """Return an element of the space."""

    def __contains__(self, x: Any) -> bool:
        return bool(self.contains(x))

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """Return if this space is equal to another space."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the space."""

    @abstractmethod
    def __hash__(self) -> int:
        """Return a hash of the space."""
